from .grad_descent import grad_descent
from .rmsprop import rmsprop
from .newton_descent import newton_descent
from .natural_grad_descent import natural_grad_descent
from .SPSA import SPSA

# from .stochastic_grad_descent import stochastic_grad_descent
from .CANS import CANS
from .iCANS import iCANS



from .knapsack import Knapsack, SlackFreeKnapsack
from .maximumcut import MaximumCut
from .minimumvertexcover import MinimumVertexCover
from .numberpartition import NumberPartition
from .shortestpath import ShortestPath
from .tsp import TSP, TSP_LP
from .portfoliooptimization import PortfolioOptimization
from .maximalindependentset import MIS
from .binpacking import BinPacking
from .vehiclerouting import VRP
from .sherrington_kirkpatrick import SK
from .kcolor import KColor
from .converters import FromDocplex2IsingModel
from .qubo import QUBO



from __future__ import annotations
from copy import deepcopy
import numpy as np

from ..backends.basebackend import QAOABaseBackend, QAOABaseBackendShotBased
from ..optimizers.logger_vqa import Logger
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)


def log_qfim_evals(logger: Logger) -> Logger:
    current_total_eval = logger.func_evals.best[0]
    current_total_eval += 1
    current_qfim_eval = logger.qfim_func_evals.best[0]
    current_qfim_eval += 1
    logger.log_variables(
        {"func_evals": current_total_eval, "qfim_func_evals": current_qfim_eval}
    )

    return logger


def qfim(
    backend_obj: QAOABaseBackend,
    params: QAOAVariationalBaseParams,
    logger: Logger,
    eta: float = 0.00000001,
):
    """
    Returns a callable qfim_fun(args) that computes the
    Quantum Fisher Information Matrix at `args` according to :
    $$[QFI]_{ij} = Re(<∂iφ|∂jφ>) − <∂iφ|φ><φ|∂jφ>$$.

    Parameters
    ----------
    params: `QAOAVariationalBaseParams`
        The QAOA parameters as a 1D array (derived from an object of one of
        the parameter classes, containing hyperparameters and variable parameters).

    eta: `float`
        The infinitesimal shift used to compute `|∂jφ>`, the partial derivative
        of the wavefunction w.r.t a parameter.

    Returns
    -------
    qfim_array:
        The quantum fisher information matrix, a 2p*2p symmetric square matrix
        with elements [QFI]_ij = Re(<∂iφ|∂jφ>) − <∂iφ|φ><φ|∂jφ>.
    """

    if isinstance(backend_obj, QAOABaseBackendShotBased):
        raise NotImplementedError(
            "QFIM computation is not currently available on shot-based"
        )

    psi = backend_obj.wavefunction(params)
    log_qfim_evals(logger)

    qfim_array = np.zeros((len(params), len(params)))

    copied_params = deepcopy(params)

    def qfim_fun(args):
        for i in range(len(args)):
            for j in range(i + 1):
                vi, vj = np.zeros(len(args)), np.zeros(len(args))
                vi[i] = eta
                vj[j] = eta

                copied_params.update_from_raw(args + vi)
                wavefunction_plus_i = np.array(backend_obj.wavefunction(copied_params))
                log_qfim_evals(logger)

                copied_params.update_from_raw(args - vi)
                wavefunction_minus_i = np.array(backend_obj.wavefunction(copied_params))
                log_qfim_evals(logger)

                copied_params.update_from_raw(args + vj)
                wavefunction_plus_j = np.array(backend_obj.wavefunction(copied_params))
                log_qfim_evals(logger)

                copied_params.update_from_raw(args - vj)
                wavefunction_minus_j = np.array(backend_obj.wavefunction(copied_params))
                log_qfim_evals(logger)

                di_psi = (wavefunction_plus_i - wavefunction_minus_i) / eta
                dj_psi = (wavefunction_plus_j - wavefunction_minus_j) / eta

                qfim_array[i][j] = np.real(np.vdot(di_psi, dj_psi)) - np.vdot(
                    di_psi, psi
                ) * np.vdot(psi, dj_psi)

                if i != j:
                    qfim_array[j][i] = qfim_array[i][j]

        return qfim_array

    return qfim_fun



from typing import Callable
from abc import ABC, abstractmethod
import numpy as np
from .rotationangle import RotationAngle


class GateApplicator(ABC):
    @abstractmethod
    def apply_gate(self, gate) -> Callable:
        """
        Apply gate to the circuit
        """
        pass


class Gate(ABC):
    def __init__(self, applicator: GateApplicator):
        self.applicator = applicator

    @abstractmethod
    def apply_gate(self, ckt):
        pass


class OneQubitGate(Gate):
    def __init__(self, applicator: GateApplicator, qubit_1: int):
        super().__init__(applicator)
        self.qubit_1 = qubit_1
        self.n_qubits = 1

    def apply_gate(self, ckt):
        return self.applicator.apply_gate(self, self.qubit_1, ckt)


class OneQubitRotationGate(OneQubitGate):
    def __init__(
        self, applicator: GateApplicator, qubit_1: int, rotation_object: RotationAngle
    ):
        super().__init__(applicator, qubit_1)
        self.rotation_object = rotation_object

    def apply_gate(self, ckt):
        return self.applicator.apply_gate(self, self.qubit_1, self.rotation_object, ckt)


class TwoQubitGate(Gate):
    def __init__(self, applicator: GateApplicator, qubit_1: int, qubit_2: int):
        super().__init__(applicator)
        self.qubit_1 = qubit_1
        self.qubit_2 = qubit_2
        self.n_qubits = 2

    def apply_gate(self, ckt):
        return self.applicator.apply_gate(self, self.qubit_1, self.qubit_2, ckt)


class TwoQubitRotationGate(TwoQubitGate):
    def __init__(
        self,
        applicator: GateApplicator,
        qubit_1: int,
        qubit_2: int,
        rotation_object: RotationAngle,
    ):
        super().__init__(applicator, qubit_1, qubit_2)
        self.rotation_object = rotation_object

    def apply_gate(self, ckt):
        return self.applicator.apply_gate(
            self, self.qubit_1, self.qubit_2, self.rotation_object, ckt
        )

    def apply_vector_gate(self, input_obj):
        input_obj.apply_rzz(
            self.qubit_1, self.qubit_2, self.rotation_object.rotation_angle
        )


class X(OneQubitGate):
    __name__ = "X"


class RZ(OneQubitRotationGate):
    __name__ = "RZ"


class RY(OneQubitRotationGate):
    __name__ = "RY"


class RX(OneQubitRotationGate):
    __name__ = "RX"


class CZ(TwoQubitGate):
    __name__ = "CZ"


class CX(TwoQubitGate):
    __name__ = "CX"


class RXX(TwoQubitRotationGate):
    __name__ = "RXX"


class RYY(TwoQubitRotationGate):
    __name__ = "RYY"


class RZZ(TwoQubitRotationGate):
    __name__ = "RZZ"


class RXY(TwoQubitRotationGate):
    __name__ = "RXY"


class RZX(TwoQubitRotationGate):
    __name__ = "RZX"


class RYZ(TwoQubitRotationGate):
    __name__ = "RYZ"


class CPHASE(TwoQubitRotationGate):
    __name__ = "CPHASE"


class RiSWAP(TwoQubitRotationGate):
    __name__ = "RiSWAP"



from __future__ import annotations
from typing import Union, Iterable
from abc import ABC
import numpy as np

from ..ansatz_constructor import QAOADescriptor


class QAOAVariationalBaseParams(ABC):
    """
    A class that initialises and keeps track of the Variational
    parameters

    Parameters
    ----------
    qaoa_descriptor: `QAOADescriptor`
        Specify the circuit parameters to construct circuit angles to be
        used for training

    Attributes
    ----------
    qaoa_descriptor: `QAOADescriptor`
    p: `int`
    cost_1q_coeffs
    cost_2q_coeffs
    mixer_1q_coeffs
    mixer_2q_coeffs
    """

    def __init__(self, qaoa_descriptor: QAOADescriptor):
        self.qaoa_descriptor = qaoa_descriptor
        self.p = self.qaoa_descriptor.p

        try:
            self.cost_1q_coeffs = qaoa_descriptor.cost_single_qubit_coeffs
            self.cost_2q_coeffs = qaoa_descriptor.cost_pair_qubit_coeffs
            self.mixer_1q_coeffs = qaoa_descriptor.mixer_single_qubit_coeffs
            self.mixer_2q_coeffs = qaoa_descriptor.mixer_pair_qubit_coeffs
        except AttributeError:
            self.cost_1q_coeffs = qaoa_descriptor.cost_hamiltonian.single_qubit_coeffs
            self.cost_2q_coeffs = qaoa_descriptor.cost_hamiltonian.pair_qubit_coeffs
            self.mixer_1q_coeffs = qaoa_descriptor.mixer_hamiltonian.single_qubit_coeffs
            self.mixer_2q_coeffs = qaoa_descriptor.mixer_hamiltonian.pair_qubit_coeffs

    def __len__(self):
        """
        Returns
        -------
        int:
            the length of the data produced by self.raw() and accepted by
            self.update_from_raw()
        """
        raise NotImplementedError()

    def __repr__(self):
        raise NotImplementedError()

    def __str__(self):
        return self.__repr__()

    @property
    def mixer_1q_angles(self) -> np.ndarray:
        """2D array with the X-rotation angles.

        1st index goes over p and the 2nd index over the qubits,
        to apply X-rotations on.
        """
        raise NotImplementedError()

    @property
    def mixer_2q_angles(self) -> np.ndarray:
        """2D array with the XX and YY-rotation angles.

        1st index goes over p and the 2nd index over the qubit pairs,
        to apply XX and YY-rotations on.
        """
        raise NotImplementedError()

    @property
    def cost_1q_angles(self) -> np.ndarray:
        """2D array with the Z-rotation angles.

        1st index goes over the p and the 2nd index over the qubits,
        to apply Z-rotations on.
        """
        raise NotImplementedError()

    @property
    def cost_2q_angles(self) -> np.ndarray:
        """2D array with ZZ-rotation angles.

        1st index goes over the p and the 2nd index over the qubit
        pairs, to apply ZZ-rotations on.
        """
        raise NotImplementedError()

    def update_from_raw(self, new_values: Union[list, np.array]):
        """
        Update all the parameters from a 1D array.

        The input has the same format as the output of ``self.raw()``.
        This is useful for ``scipy.optimize.minimize`` which expects
        the parameters that need to be optimized to be a 1D array.

        Parameters
        ----------
        new_values: `Union[list, np.array]`
            A 1D array with the new parameters. Must have length  ``len(self)``
            and the ordering of the flattend ``parameters`` in ``__init__()``.

        """
        raise NotImplementedError()

    def raw(self) -> np.ndarray:
        """
        Return the parameters in a 1D array.

        This 1D array is needed by ``scipy.optimize.minimize`` which expects
        the parameters that need to be optimized to be a 1D array.

        Returns
        -------
        np.array:
            The parameters in a 1D array. Has the same output format as the
            expected input of ``self.update_from_raw``. Hence corresponds to
            the flattened `parameters` in `__init__()`

        """
        raise NotImplementedError()

    def update_from_dict(self, new_values: dict):
        """
        Update all the parameters from a dictionary.

        The input has the same format as the output of ``self.asdict()``.

        Parameters
        ----------
        new_values: `dict`
            A dictionary with the new parameters. Must have the same keys as
            the output of ``self.asdict()``.

        """

        assert isinstance(new_values, dict), f"Expected dict, got {type(new_values)}"

        for key, value in new_values.items():
            if key not in self.asdict().keys():
                raise KeyError(
                    f"'{key}' not in {self.__class__.__name__}, expected keys: {list(self.asdict().keys())}"
                )
            else:
                if getattr(self, key).shape != np.array(value).shape:
                    raise ValueError(
                        f"Shape of '{key}' does not match. Expected shape {getattr(self, key).shape}, got {np.array(value).shape}."
                    )

        raw_params = []
        for key, value in self.asdict().items():
            if key in new_values.keys():
                raw_params += list(np.array(new_values[key]).flatten())
            else:
                raw_params += list(np.array(value).flatten())

        self.update_from_raw(raw_params)

    def asdict(self) -> dict:
        """
        Return the parameters as a dictionary.

        Returns
        -------
        dict:
            The parameters as a dictionary. Has the same output format as the
            expected input of ``self.update_from_dict``.

        """
        return {k[2:]: v for k, v in self.__dict__.items() if k[0:2] == "__"}

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, time: float = None
    ):
        """Alternative to ``__init__`` that already fills ``parameters``.

        Calculate initial parameters from register, terms, weights
        (specifiying a Hamiltonian), corresponding to a linear ramp
        annealing schedule and return a ``QAOAVariationalBaseParams`` object.

        Parameters
        ----------
        qaoa_descriptor: `QAOADescriptor`
            QAOADescriptor object containing information about terms,weights,register and p

        time: `float`
            Total annealing time. Defaults to ``0.7*p``.

        Returns
        -------
        QAOAVariationalBaseParams:
            The initial parameters for a linear ramp for ``hamiltonian``.

        """
        raise NotImplementedError()

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, seed: int = None):
        """
        Initialise parameters randomly

        Parameters
        ----------
        qaoa_descriptor: `QAOADescriptor`
            QAOADescriptor object containing information about terms,
            weights, register and p.

        seed: `int`
                Use a fixed seed for reproducible random numbers

        Returns
        -------
        QAOAVariationalBaseParams:
            Randomly initialiased parameters
        """
        raise NotImplementedError()

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor):
        """
        Alternative to ``__init__`` that only takes ``qaoa_descriptor`` and
        fills ``parameters`` via ``np.empty``

        Parameters
        ----------
        qaoa_descriptor: `QAOADescriptor`
            QAOADescriptor object containing information about terms,weights,register and p

        Returns
        -------
        QAOAVariationalBaseParams:
            A Parameter object with the parameters filled by ``np.empty``
        """
        raise NotImplementedError()

    @classmethod
    def from_other_parameters(cls, params):
        """Alternative to ``__init__`` that takes parameters with less degrees
        of freedom as the input.

        Parameters
        ----------
        params: `QAOAVaritionalBaseParams`
            The input parameters object to construct the new parameters object from.
        Returns
        -------
        QAOAVariationalBaseParams:
            The converted paramters s.t. all the rotation angles of the in
            and output parameters are the same.
        """
        from . import converter

        return converter(params, cls)

    def raw_rotation_angles(self) -> np.ndarray:
        """
        Flat array of the rotation angles for the memory map for the
        parametric circuit.

        Returns
        -------
        np.array:
            Returns all single rotation angles in the ordering
            ``(x_rotation_angles, gamma_singles, zz_rotation_angles)`` where
            ``x_rotation_angles = (beta_q0_t0, beta_q1_t0, ... , beta_qn_tp)``
            and the same for ``z_rotation_angles`` and ``zz_rotation_angles``

        """
        raw_data = np.concatenate(
            (
                self.mixer_1q_angles.flatten(),
                self.mixer_2q_angles.flatten(),
                self.cost_1q_angles.flatten(),
                self.cost_1q_angles.flatten(),
            )
        )
        return raw_data

    def plot(self, ax=None, **kwargs):
        """
        Plots ``self`` in a sensible way to the canvas ``ax``, if provided.

        Parameters
        ----------
        ax: `matplotlib.axes._subplots.AxesSubplot`
                The canvas to plot itself on
        kwargs:
                All remaining keyword arguments are passed forward to the plot
                function

        """
        raise NotImplementedError()


class QAOAParameterIterator:
    """An iterator to sweep one parameter over a range in a QAOAParameter object.

    Parameters
    ----------
    qaoa_params:
        The initial QAOA parameters, where one of them is swept over
    the_parameter:
        A string specifying, which parameter should be varied. It has to be
        of the form ``<attr_name>[i]`` where ``<attr_name>`` is the name
        of the _internal_ list and ``i`` the index, at which it sits. E.g.
        if ``qaoa_params`` is of type ``AnnealingParams``
        and  we want to vary over the second timestep, it is
        ``the_parameter = "times[1]"``.
    the_range:
        The range, that ``the_parameter`` should be varied over

    Todo
    ----
    - Add checks, that the number of indices in ``the_parameter`` matches
      the dimensions of ``the_parameter``
    - Add checks, that the index is not too large

    Example
    -------
    Assume qaoa_params is of type ``StandardWithBiasParams`` and
    has `p >= 2`. Then the following code produces a loop that
    sweeps ``gammas_singles[1]`` over the range ``(0, 1)`` in 4 layers:

    .. code-block:: python

        the_range = np.arange(0, 1, 0.4)
        the_parameter = "gammas_singles[1]"
        param_iterator = QAOAParameterIterator(qaoa_params, the_parameter, the_range)
        for params in param_iterator:
            # do what ever needs to be done.
            # we have type(params) == type(qaoa_params)
    """

    def __init__(
        self,
        variational_params: QAOAVariationalBaseParams,
        the_parameter: str,
        the_range: Iterable[float],
    ):
        """See class documentation for details"""
        self.params = variational_params
        self.iterator = iter(the_range)
        self.the_parameter, *indices = the_parameter.split("[")
        indices = [i.replace("]", "") for i in indices]
        if len(indices) == 1:
            self.index0 = int(indices[0])
            self.index1 = False
        elif len(indices) == 2:
            self.index0 = int(indices[0])
            self.index1 = int(indices[1])
        else:
            raise ValueError("the_parameter has to many indices")

    def __iter__(self):
        return self

    def __next__(self):
        # get next value from the_range
        value = next(self.iterator)

        # 2d list or 1d list?
        if self.index1 is not False:
            getattr(self.params, self.the_parameter)[self.index0][self.index1] = value
        else:
            getattr(self.params, self.the_parameter)[self.index0] = value

        return self.params



import numpy as np
from scipy.optimize import OptimizeResult


def grad_descent(
    fun,
    x0,
    args=(),
    maxfev=None,
    stepsize=0.01,
    maxiter=100,
    tol=10 ** (-6),
    jac=None,
    callback=None,
    **options
):
    """
    Minimize a function `fun` using gradient descent. scipy.optimize.minimize
    compatible implementation of gradient descent for `method` == 'vgd'.

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the
        difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    """

    bestx = x0
    besty = fun(x0)
    funcalls = 1  # tracks no. of function evals.
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # compute gradient descent step
        testx = testx - stepsize * jac(testx)
        testy = np.real(fun(testx, *args))

        if np.abs(besty - testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(
        fun=besty, x=bestx, nit=niter, nfev=funcalls, success=(niter > 1)
    )



import unittest
import numpy as np

from openqaoa.backends.qaoa_analytical_sim import QAOABackendAnalyticalSimulator
from openqaoa.algorithms import QAOA, RQAOA
from openqaoa.problems import MaximumCut
from openqaoa.backends.qaoa_device import create_device
from openqaoa.utilities import (
    X_mixer_hamiltonian,
    XY_mixer_hamiltonian,
    ring_of_disagrees,
    random_k_regular_graph,
)
from openqaoa.qaoa_components import QAOADescriptor, QAOAVariationalStandardParams

"""
A set of tests for the analytical simulator backend which computes the energy of a given quantum circuit as a function of beta and gamma.
"""


def Disagrees_SetUp(n_qubits, p, mixer_hamil, betas, gammas):
    """
    Helper function for the tests below
    """

    register = range(n_qubits)
    cost_hamil = ring_of_disagrees(register)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p)
    variational_params_std = QAOAVariationalStandardParams(
        qaoa_descriptor, betas, gammas
    )
    # Get the part of the Hamiltonian proportional to the identity

    return register, cost_hamil, qaoa_descriptor, variational_params_std


class TestingQAOABackendAnalyticalSimulator(unittest.TestCase):
    """
    Unittest based testing of QAOABackendAnalyticalSimulator
    """

    def test_expectation(self):
        """
        Testing if the analytical formula returns the energy expectation value for a given beta and gamma on an easy problem (Ring of disagree).
        """
        n_qubits = 8
        p = 1
        mixer_hamil = X_mixer_hamiltonian(n_qubits)
        betas = [np.pi / 8]
        gammas = [np.pi / 4]
        register, cost_hamil, qaoa_descriptor, variate_params = Disagrees_SetUp(
            n_qubits, p, mixer_hamil, betas, gammas
        )

        backend_analytical = QAOABackendAnalyticalSimulator(qaoa_descriptor)
        exp_val = backend_analytical.expectation(variate_params)

        # Check correct expecation value
        assert np.isclose(exp_val, -6)

    def test_p_not_1_fails(self):
        """
        Testing if the analytical backend fails if the number of layers, p, is different than 1.
        """
        n_qubits = 8
        p = 2
        mixer_hamil = X_mixer_hamiltonian(n_qubits)
        betas = [np.pi / 8, np.pi / 8]
        gammas = [np.pi / 4, np.pi / 4]
        register, cost_hamil, qaoa_descriptor, variate_params = Disagrees_SetUp(
            n_qubits, p, mixer_hamil, betas, gammas
        )
        exception = False
        try:
            backend_analytical = QAOABackendAnalyticalSimulator(qaoa_descriptor)
        except:
            exception = True

        assert exception, "p != 1 didn't fail."

    def test_different_mixer_fails(self):
        """
        Testing if the analytical backend fails if the mixer hamiltonian is different than X.
        """
        exception = False
        n_qubits = 8
        p = 1
        mixer_hamil = XY_mixer_hamiltonian(n_qubits)
        betas = [np.pi / 8]
        gammas = [np.pi / 4]
        register, cost_hamil, qaoa_descriptor, variate_params = Disagrees_SetUp(
            n_qubits, p, mixer_hamil, betas, gammas
        )
        try:
            backend_analytical = QAOABackendAnalyticalSimulator(qaoa_descriptor)
        except:
            exception = True

        assert exception, "XY mixer Hamiltonian didn't fail."

    def test_not_standard_params_fails(self):
        """
        Testing if the analytical backend fails if the parametrization of the circuit is extended or fourier.
        """
        # Create a 3-regular weighted graph and a qubo problem
        g = random_k_regular_graph(
            degree=3, nodes=range(8), seed=2642, weighted=True, biases=False
        )
        maxcut_qubo = MaximumCut(g).qubo

        # Testing for Extended params
        exception = False
        q = QAOA()
        q.set_circuit_properties(
            p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
        )
        q.set_classical_optimizer(method="nelder-mead", maxiter=10)
        analytical_device = create_device(location="local", name="analytical_simulator")
        q.set_device(analytical_device)
        q.compile(maxcut_qubo)
        try:
            q.optimize()
        except:
            exception = True
        assert exception, "Extended params didn't fail"

        # Testing for Fourier params
        q = QAOA()
        q.set_device(analytical_device)
        q.set_circuit_properties(
            p=1, param_type="fourier", init_type="rand", mixer_hamiltonian="x"
        )
        exception = False
        try:
            q.optimize()
        except:
            exception = True
        assert exception, "Fourier params didn't fail"

    def test_end_to_end_rqaoa(self):
        """
        Testing the whole rqaoa workflow if the device is set to 'analytical_simulator'.
        """
        # Create a 3-regular weighted graph and a qubo problem
        g = random_k_regular_graph(
            degree=3, nodes=range(8), seed=2642, weighted=True, biases=False
        )
        maxcut_qubo = MaximumCut(g).qubo

        # Define the RQAOA object and set its params
        r = RQAOA()
        r.set_rqaoa_parameters(steps=1, n_cutoff=3)
        r.set_circuit_properties(
            p=1,
            init_type="custom",
            variational_params_dict={"betas": [0.26], "gammas": [0.42]},
            mixer_hamiltonian="x",
        )

        # Define the device to be the analytical simulator
        device = create_device(location="local", name="analytical_simulator")
        r.set_device(device)

        # Set the classical method used to optimize over QAOA angles and its properties
        r.set_classical_optimizer(
            method="rmsprop",
            optimizer_options={"stepsize": 10 ** (-10)},
            tol=10 ** (-1),
            maxfev=1,
            maxiter=1,
            jac="finite_difference",
            optimization_progress=False,
            cost_progress=False,
            parameter_log=False,
        )

        # Compile and optimize the problem instance on RQAOA
        r.compile(maxcut_qubo)
        r.optimize()

        opt_results = r.result
        opt_solution = opt_results["solution"]
        opt_solution_string = list(opt_solution.keys())[0]

        assert opt_solution_string == "01011010"

    def test_exact_solution(self):
        """
        Testing the exact solution which is a property of every backend.

        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        correct_energy = -8
        correct_config = [0, 1, 0, 1, 0, 1, 0, 1]

        # The tests pass regardless of the value of betas and gammas is this correct?
        betas = [np.pi / 8]
        gammas = [np.pi / 4]

        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_analytical = QAOABackendAnalyticalSimulator(qaoa_descriptor)
        # exact solution is defined as the property of the cost function
        energy_vec, config_vec = backend_analytical.exact_solution

        assert np.isclose(energy_vec, correct_energy)

        config_vec = [config.tolist() for config in config_vec]

        assert correct_config in config_vec


if __name__ == "__main__":
    unittest.main()



import unittest
import itertools
import networkx as nw
import numpy as np

from openqaoa import QAOA, RQAOA
from openqaoa.algorithms import QAOAResult, RQAOAResult
from openqaoa.backends.qaoa_backend import (
    DEVICE_NAME_TO_OBJECT_MAPPER,
    DEVICE_ACCESS_OBJECT_MAPPER,
)
from openqaoa.backends import create_device
from openqaoa.backends.basebackend import QAOABaseBackendStatevector
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS
from openqaoa.problems import MinimumVertexCover, QUBO, MaximumCut
from openqaoa.qaoa_components import Hamiltonian


def _compare_qaoa_results(dict_old, dict_new, bool_cmplx_str):
    for key in dict_old.keys():
        if key == "cost_hamiltonian":  ## CHECK WHAT DO WITH THIS
            pass
        elif key == "_QAOAResult__type_backend":
            if issubclass(dict_old[key], QAOABaseBackendStatevector):
                assert (
                    dict_new[key] == QAOABaseBackendStatevector
                ), "Type of backend is not correct, complex_to_string = {}".format(
                    bool_cmplx_str
                )
            else:
                assert (
                    dict_new[key] == ""
                ), "Type of backend should be empty string, complex_to_string = {}".format(
                    bool_cmplx_str
                )
        elif key == "optimized":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    assert np.all(
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same, complex_to_string = {}".format(
                        bool_cmplx_str
                    )
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same, complex_to_string = {}".format(
                        bool_cmplx_str
                    )
        elif key == "intermediate":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    for step in range(len(dict_old[key][key2])):
                        assert np.all(
                            dict_old[key][key2][step] == dict_new[key][key2][step]
                        ), "Intermediate params are not the same, complex_to_string = {}".format(
                            bool_cmplx_str
                        )
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Intermediate params are not the same, complex_to_string = {}".format(
                        bool_cmplx_str
                    )
        else:
            assert (
                dict_old[key] == dict_new[key]
            ), f"{key} is not the same, complex_to_string = {bool_cmplx_str}"


def _test_keys_in_dict(obj, expected_keys):
    """
    private function to test the keys. It recursively tests the keys of the nested dictionaries, or lists of dictionaries
    """

    if isinstance(obj, dict):
        for key in obj:
            if key in expected_keys.keys():
                expected_keys[key] = True

            if isinstance(obj[key], dict):
                _test_keys_in_dict(obj[key], expected_keys)
            elif isinstance(obj[key], list):
                for item in obj[key]:
                    _test_keys_in_dict(item, expected_keys)
    elif isinstance(obj, list):
        for item in obj:
            _test_keys_in_dict(item, expected_keys)


class TestingResultOutputs(unittest.TestCase):

    """
    Test the Results Output after an optimization loop
    """

    def test_flags_result_outputs_workflow(self):
        """
        Run an optimization problem for 5 iterations.
        Should expect certain fields of the results output to be filled based
        on some of the users inputs. (Default settings)
        Can be checked for cobyla.

        Check for all available supported local backends.
        """

        g = nw.circulant_graph(3, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        choice_combination = list(
            itertools.product([True, False], [True, False], [True, False])
        )
        recorded_evals = [0, 5]

        for device_name in SUPPORTED_LOCAL_SIMULATORS:
            for each_choice in choice_combination:
                q = QAOA()
                q.set_classical_optimizer(
                    method="cobyla",
                    parameter_log=each_choice[0],
                    cost_progress=each_choice[1],
                    optimization_progress=each_choice[2],
                    maxiter=5,
                )
                device = create_device("local", device_name)
                print(device.device_name)
                q.set_device(device)
                q.compile(vc)
                q.optimize()
                self.assertEqual(
                    recorded_evals[each_choice[0]], len(q.result.intermediate["angles"])
                )
                self.assertEqual(
                    recorded_evals[each_choice[1]], len(q.result.intermediate["cost"])
                )
                self.assertEqual(
                    recorded_evals[each_choice[2]],
                    len(q.result.intermediate["measurement_outcomes"]),
                )

    def test_qaoa_result_asdict(self):
        """
        Test the qaoa result.asdict method
        """

        # run the QAOA
        qaoa = QAOA()
        qaoa.compile(problem=QUBO.random_instance(n=8))
        qaoa.optimize()

        # get dict
        results_dict = qaoa.result.asdict()

        # list of expected keys
        expected_keys = [
            "method",
            "cost_hamiltonian",
            "n_qubits",
            "terms",
            "qubit_indices",
            "pauli_str",
            "phase",
            "coeffs",
            "constant",
            "qubits_pairs",
            "qubits_singles",
            "single_qubit_coeffs",
            "pair_qubit_coeffs",
            "evals",
            "number_of_evals",
            "jac_evals",
            "qfim_evals",
            "most_probable_states",
            "solutions_bitstrings",
            "bitstring_energy",
            "intermediate",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
            "optimized",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
        ]

        # we append all the keys that we find in rqaoa.results, so if we introduce a new key, we will know that we need to update the result.asdict method
        for key in vars(qaoa.result).keys():
            if not key in expected_keys and not "_QAOAResult__" in key:
                expected_keys.append(key)

        # create a dictionary with all the expected keys and set them to False
        expected_keys_dict = {item: False for item in expected_keys}

        # test the keys, it will set the keys to True if they are found
        _test_keys_in_dict(results_dict, expected_keys_dict)

        # Check if the dictionary has all the expected keys
        for key, value in expected_keys_dict.items():
            assert (
                value == True
            ), f"Key {key} was not found in the dictionary of the QAOA Result class."

        ## now we repeat the same test but we do not include the cost hamiltonian

        # get dict without cost hamiltonian
        results_dict = qaoa.result.asdict(keep_cost_hamiltonian=False)

        # expected keys
        expected_keys_dict = {item: False for item in expected_keys}
        expected_keys_not_in_dict = [
            "cost_hamiltonian",
            "n_qubits",
            "terms",
            "qubit_indices",
            "pauli_str",
            "phase",
            "coeffs",
            "constant",
            "qubits_pairs",
            "qubits_singles",
            "single_qubit_coeffs",
            "pair_qubit_coeffs",
        ]

        # test the keys, it will set the keys to True if they are found, except the ones that were not included which should be those in expected_keys_not_in_dict
        _test_keys_in_dict(results_dict, expected_keys_dict)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys_dict.items():
            if not key in expected_keys_not_in_dict:
                assert (
                    value == True
                ), f"Key {key} was not found in the dictionary of the RQAOAResult class."
            else:
                assert (
                    value == False
                ), f"Key {key} was found in the dictionary of the RQAOAResult class, but it should not have been."

        ## now we repeat the same test but we do not include some keys

        # get dict without some values
        results_dict = qaoa.result.asdict(
            exclude_keys=["solutions_bitstrings", "method"]
        )

        # expected keys
        expected_keys_dict = {item: False for item in expected_keys}
        expected_keys_not_in_dict = ["solutions_bitstrings", "method"]

        # test the keys, it will set the keys to True if they are found, except the ones that were not included which should be those in expected_keys_not_in_dict
        _test_keys_in_dict(results_dict, expected_keys_dict)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys_dict.items():
            if not key in expected_keys_not_in_dict:
                assert (
                    value == True
                ), f"Key {key} was not found in the dictionary of the RQAOAResult class."
            else:
                assert (
                    value == False
                ), f"Key {key} was found in the dictionary of the RQAOAResult class, but it should not have been."

        """
        to get the list of expected keys, run the following code:

            def get_keys(obj, list_keys):
                if isinstance(obj, dict):
                    for key in obj:
                        if not key in list_keys: list_keys.append(key)

                        if isinstance(obj[key], dict):
                            get_keys(obj[key], list_keys)
                        elif isinstance(obj[key], list):
                            for item in obj[key]:
                                get_keys(item, list_keys)
                elif isinstance(obj, list):
                    for item in obj:
                        get_keys(item, list_keys)

            expected_keys = []
            get_keys(rqaoa.results.asdict(), expected_keys)
            print(expected_keys)
        """

    # test eval_number
    def test_qaoa_result_eval_number(self):
        """
        Test the eval_number method for the QAOA result class
        """

        for method in ["cobyla", "spsa", "vgd", "newton", "natural_grad_descent"]:
            # run the QAOA and get the results
            q = QAOA()
            q.set_classical_optimizer(
                maxiter=15,
                method=method,
                jac="finite_difference",
                hess="finite_difference",
            )
            q.compile(problem=QUBO.random_instance(n=8))
            q.optimize()

            # test the eval_number method
            assert (
                q.result.intermediate["cost"].index(min(q.result.intermediate["cost"]))
                + 1
                == q.result.optimized["eval_number"]
            ), "optimized eval_number does not return the correct number of the optimized evaluation, when using {} method".format(
                method
            )

    def test_qaoa_results_from_dict(self):
        """
        test loading the QAOA Result object from a dictionary
        methods: from_dict
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run qaoa with different devices, and save the objects in a list
        qaoas = []
        for device in [
            create_device(location="local", name=each_device_name)
            for each_device_name in SUPPORTED_LOCAL_SIMULATORS
            if each_device_name != "analytical_simulator"
        ]:
            q = QAOA()
            q.set_device(device)
            q.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            q.set_backend_properties(prepend_state=None, append_state=None)
            q.set_classical_optimizer(maxiter=10, optimization_progress=True)

            q.compile(maxcut_qubo)

            q.optimize()

            qaoas.append(q)

        for q in qaoas:
            results = q.result
            for bool_cmplx_str in [True, False]:
                results_from_dict = QAOAResult.from_dict(
                    results.asdict(complex_to_string=bool_cmplx_str)
                )

                # assert that results_from_dict is intance of Result
                assert isinstance(
                    results_from_dict, QAOAResult
                ), "results_from_dict is not an instance of Result"

                _compare_qaoa_results(
                    results.__dict__, results_from_dict.__dict__, bool_cmplx_str
                )

    def test_qaoa_results_calculate_statistics_full(self):
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        qaoas = []
        for device_name in SUPPORTED_LOCAL_SIMULATORS:
            if device_name == "analytical_simulator":
                continue

            qaoa = QAOA()

            qaoa.set_device(create_device("local", device_name))
            qaoa.set_classical_optimizer(optimization_progress=True)

            qaoa.compile(maxcut_qubo)
            qaoa.optimize()

            qaoas.append(qaoa)

        for qaoa in qaoas:
            result = qaoa.result.calculate_statistics(include_intermediate=True)
            result_optimized = result["optimized"]
            result_intermediate = result["intermediate"]
            optimized_sorted, optimized_mean, optimized_std_deviation = (
                list(result_optimized["sorted"].values()),
                result_optimized["mean"],
                result_optimized["std_deviation"],
            )

            for ri in result_intermediate:
                intermediate_sorted, intermediate_mean, intermediate_std_deviation = (
                    list(ri["sorted"].values()),
                    ri["mean"],
                    ri["std_deviation"],
                )

                assert all(
                    intermediate_sorted[i] >= intermediate_sorted[i + 1]
                    for i in range(len(intermediate_sorted) - 1)
                ), "Counts in descending order."
                assert intermediate_mean > 0, "Mean is greater then zero."
                assert (
                    intermediate_std_deviation > 0
                ), "Standard deviation is greater than zero."

            assert all(
                optimized_sorted[i] >= optimized_sorted[i + 1]
                for i in range(len(optimized_sorted) - 1)
            ), "Counts in descending order."
            assert optimized_mean > 0, "Mean is greater then zero."
            assert (
                optimized_std_deviation > 0
            ), "Standard deviation is greater than zero."

    def test_qaoa_results_calculate_statistics_without_intermediate(self):
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        qaoas = []
        for device_name in SUPPORTED_LOCAL_SIMULATORS:
            if device_name == "analytical_simulator":
                continue

            qaoa = QAOA()
            qaoa.set_device(create_device("local", device_name))
            qaoa.set_classical_optimizer(optimization_progress=False)

            qaoa.compile(maxcut_qubo)
            qaoa.optimize()

            qaoas.append(qaoa)

        for qaoa in qaoas:
            result = qaoa.result.calculate_statistics(include_intermediate=False)
            result_optimized = result["optimized"]
            optimized_sorted, optimized_mean, optimized_std_deviation = (
                list(result_optimized["sorted"].values()),
                result_optimized["mean"],
                result_optimized["std_deviation"],
            )

            assert (
                result["intermediate"] == []
            ), "Statistics for intermediate measurements are empty."
            assert all(
                optimized_sorted[i] >= optimized_sorted[i + 1]
                for i in range(len(optimized_sorted) - 1)
            ), "Counts in descending order."
            assert optimized_mean > 0, "Mean is greater then zero."
            assert (
                optimized_std_deviation > 0
            ), "Standard deviation is greater than zero."

    def test_qaoa_results_calculate_statistics_raise_value_error(self):
        """
        The method raise an error if the user requires statistics on all
        intermediate measurement outcomes but hasn't specified saving them during optimization.
        """
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        qaoa = QAOA()
        qaoa.set_device(create_device(location="local", name="vectorized"))
        qaoa.set_classical_optimizer(optimization_progress=False)

        qaoa.compile(maxcut_qubo)
        qaoa.optimize()

        try:
            qaoa.result.calculate_statistics(include_intermediate=True)
        except ValueError as e:
            self.assertEqual(
                str(e),
                "The underlying QAOA object does not seem to have any intermediate measurement result. Please, consider saving intermediate measurements during optimization by setting `optimization_progress=True` in your workflow.",
            )


class TestingRQAOAResultOutputs(unittest.TestCase):
    """
    Test the  Results Output after a full RQAOA loop
    """

    def __run_rqaoa(
        self,
        type="custom",
        eliminations=1,
        p=1,
        param_type="standard",
        mixer="x",
        method="cobyla",
        maxiter=15,
        name_device="vectorized",
    ):
        """
        private function to run the RQAOA
        """

        n_qubits = 6
        n_cutoff = 3
        g = nw.circulant_graph(n_qubits, [1])
        problem = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        r = RQAOA()
        qiskit_device = create_device(location="local", name=name_device)
        r.set_device(qiskit_device)
        if type == "adaptive":
            r.set_rqaoa_parameters(
                n_cutoff=n_cutoff, n_max=eliminations, rqaoa_type=type
            )
        else:
            r.set_rqaoa_parameters(
                n_cutoff=n_cutoff, steps=eliminations, rqaoa_type=type
            )
        r.set_circuit_properties(p=p, param_type=param_type, mixer_hamiltonian=mixer)
        r.set_backend_properties(prepend_state=None, append_state=None)
        r.set_classical_optimizer(
            method=method,
            maxiter=maxiter,
            optimization_progress=True,
            cost_progress=True,
            parameter_log=True,
        )
        r.compile(problem)
        r.optimize()

        return r.result

    def test_rqaoa_result_outputs(self):
        """
        Test the result outputs for the RQAOA class
        """

        n_qubits = 6
        n_cutoff = 3

        # Test for the standard RQAOA
        results = self.__run_rqaoa()
        assert isinstance(
            results, RQAOAResult
        ), "Results of RQAOA are not of type RQAOAResult"
        for key in results["solution"].keys():
            assert len(key) == n_qubits, "Number of qubits solution is not correct"
        assert isinstance(results["classical_output"]["minimum_energy"], float)
        assert isinstance(results["classical_output"]["optimal_states"], list)
        for rule_list in results["elimination_rules"]:
            for rule in rule_list:
                assert isinstance(
                    rule, dict
                ), "Elimination rule item is not a dictionary"
        assert isinstance(results["schedule"], list), "Schedule is not a list"
        assert (
            sum(results["schedule"]) + n_cutoff == n_qubits
        ), "Schedule is not correct"
        for step in results["intermediate_steps"]:
            assert isinstance(step["problem"], QUBO), "problem is not of type QUBO"
            assert isinstance(
                step["qaoa_results"], QAOAResult
            ), "QAOA_results is not of type QAOA Results"
            assert isinstance(
                step["exp_vals_z"], np.ndarray
            ), "exp_vals_z is not of type numpy array"
            assert isinstance(
                step["corr_matrix"], np.ndarray
            ), "corr_matrix is not of type numpy array"
        assert isinstance(
            results["number_steps"], int
        ), "Number of steps is not an integer"

        # Test for the adaptive RQAOA
        results = self.__run_rqaoa(type="adaptive")
        assert isinstance(
            results, RQAOAResult
        ), "Results of RQAOA are not of type RQAOAResult"
        for key in results["solution"].keys():
            assert len(key) == n_qubits, "Number of qubits solution is not correct"
        assert isinstance(results["classical_output"]["minimum_energy"], float)
        assert isinstance(results["classical_output"]["optimal_states"], list)
        for rule_list in results["elimination_rules"]:
            for rule in rule_list:
                assert isinstance(
                    rule, dict
                ), "Elimination rule item is not a dictionary"
        assert isinstance(results["schedule"], list), "Schedule is not a list"
        assert (
            sum(results["schedule"]) + n_cutoff == n_qubits
        ), "Schedule is not correct"
        for step in results["intermediate_steps"]:
            assert isinstance(step["problem"], QUBO), "QUBO is not of type QUBO"
            assert isinstance(
                step["qaoa_results"], QAOAResult
            ), "QAOA_results is not of type QAOA Results"
            assert isinstance(
                step["exp_vals_z"], np.ndarray
            ), "exp_vals_z is not of type numpy array"
            assert isinstance(
                step["corr_matrix"], np.ndarray
            ), "corr_matrix is not of type numpy array"
        assert isinstance(
            results["number_steps"], int
        ), "Number of steps is not an integer"

    def test_rqaoa_result_methods_steps(self):
        """
        Test the methods for the RQAOAResult class for the steps
        """

        # run the RQAOA
        results = self.__run_rqaoa()

        # test the solution method
        assert (
            results.get_solution() == results["solution"]
        ), "get_solution method is not correct"

        # test the methods for the intermediate steps
        for i in range(results["number_steps"]):
            # methods for intermediate qaao results
            assert (
                results.get_qaoa_results(i)
                == results["intermediate_steps"][i]["qaoa_results"]
            ), "get_qaoa_results method is not correct"
            assert (
                results.get_qaoa_optimized_angles(i)
                == results.get_qaoa_results(i).optimized["angles"]
            ), "get_qaoa_optimized_angles method is not correct"

            # methods for intermediate qubo
            assert (
                results.get_problem(i) == results["intermediate_steps"][i]["problem"]
            ), "get_qubo method is not correct"
            assert isinstance(
                results.get_hamiltonian(i), Hamiltonian
            ), "get_hamiltonian method is not correct"

            # methods for intermediate exp_vals_z and corr_matrix
            assert (
                results.get_exp_vals_z(i)
                is results["intermediate_steps"][i]["exp_vals_z"]
            ), "get_exp_vals_z method is not correct"
            assert (
                results.get_corr_matrix(i)
                is results["intermediate_steps"][i]["corr_matrix"]
            ), "get_corr_matrix method is not correct"

    def test_rqaoa_result_plot_corr_matrix(self):
        """
        Test the plot_corr_matrix method for the RQAOAResult class
        """

        # run the RQAOA
        results = self.__run_rqaoa()

        # test the plot_corr_matrix method
        for i in range(results["number_steps"]):
            results.plot_corr_matrix(step=i)

    def test_rqaoa_result_asdict(self):
        """
        Test the plot_exp_vals_z method for the RQAOAResult class
        """

        # run the RQAOA
        results = self.__run_rqaoa()

        # get dict
        results_dict = results.asdict()

        # create a list of expected keys
        expected_keys = [
            "solution",
            "classical_output",
            "minimum_energy",
            "optimal_states",
            "elimination_rules",
            "singlet",
            "bias",
            "pair",
            "correlation",
            "schedule",
            "intermediate_steps",
            "problem",
            "terms",
            "weights",
            "constant",
            "n",
            "qaoa_results",
            "method",
            "cost_hamiltonian",
            "n_qubits",
            "qubit_indices",
            "pauli_str",
            "phase",
            "coeffs",
            "qubits_pairs",
            "qubits_singles",
            "single_qubit_coeffs",
            "pair_qubit_coeffs",
            "evals",
            "number_of_evals",
            "jac_evals",
            "qfim_evals",
            "most_probable_states",
            "solutions_bitstrings",
            "bitstring_energy",
            "intermediate",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
            "optimized",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
            "exp_vals_z",
            "corr_matrix",
            "number_steps",
        ]

        # we append all the keys that we find in rqaoa.results, so if we introduce a new key, we will know that we need to update the result.asdict method
        for key in results.keys():
            if not key in expected_keys:
                expected_keys.append(key)
        for key in results["intermediate_steps"][0].keys():
            if not key in expected_keys:
                expected_keys.append(key)

        # dictionary with all the expected keys and set them to False
        expected_keys = {item: False for item in expected_keys}

        # test the keys, it will set the keys to True if they are found
        _test_keys_in_dict(results_dict, expected_keys)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys.items():
            assert (
                value == True
            ), f"Key {key} was not found in the dictionary of the RQAOAResult class."

        ## now we repeat the same test but we do not include some keys

        # get dict without some values
        results_dict = results.asdict(exclude_keys=["solutions_bitstrings", "method"])

        # expected keys
        expected_keys_dict = {item: False for item in expected_keys}
        expected_keys_not_in_dict = ["solutions_bitstrings", "method"]

        # test the keys, it will set the keys to True if they are found, except the ones that were not included which should be those in expected_keys_not_in_dict
        _test_keys_in_dict(results_dict, expected_keys_dict)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys_dict.items():
            if not key in expected_keys_not_in_dict:
                assert (
                    value == True
                ), f"Key {key} was not found in the dictionary of the RQAOAResult class."
            else:
                assert (
                    value == False
                ), f"Key {key} was found in the dictionary of the RQAOAResult class, but it should not have been."

        """
        to get the list of expected keys, run the following code:

            def get_keys(obj, list_keys):
                if isinstance(obj, dict):
                    for key in obj:
                        if not key in list_keys: list_keys.append(key)

                        if isinstance(obj[key], dict):
                            get_keys(obj[key], list_keys)
                        elif isinstance(obj[key], list):
                            for item in obj[key]:
                                get_keys(item, list_keys)
                elif isinstance(obj, list):
                    for item in obj:
                        get_keys(item, list_keys)

            expected_keys = []
            get_keys(rqaoa.results.asdict(), expected_keys)
            print(expected_keys)
        """

    def test_rqaoa_results_from_dict(self):
        """
        test loading the QAOA Result object from a dictionary
        methods: from_dict
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run rqaoa with different devices, and save the objcets in a list
        rqaoas = []

        for device in [
            create_device(location="local", name=each_device_name)
            for each_device_name in SUPPORTED_LOCAL_SIMULATORS
            if each_device_name != "analytical_simulator"
        ]:
            r = RQAOA()
            r.set_device(device)
            r.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            r.set_backend_properties(prepend_state=None, append_state=None)
            r.set_classical_optimizer(maxiter=10, optimization_progress=True)

            r.compile(maxcut_qubo)

            r.optimize()

            rqaoas.append(r)

        # for each rqaoa object, we check that we can create a new results object from the dictionary of the old one
        for r in rqaoas:
            new_results = RQAOAResult.from_dict(r.result.asdict())
            old_results = r.result

            # assert that new_results is an instance of RQAOAResult
            assert isinstance(
                new_results, RQAOAResult
            ), "new_results is not an instance of RQAOAResult"

            for key in old_results:
                if key == "intermediate_steps":
                    for i in range(len(old_results[key])):
                        for key2 in old_results[key][i]:
                            if key2 == "problem":
                                assert (
                                    old_results[key][i][key2].asdict()
                                    == new_results[key][i][key2].asdict()
                                ), f"{key2} is not the same"
                            elif key2 == "qaoa_results":
                                _compare_qaoa_results(
                                    old_results[key][i][key2].asdict(),
                                    new_results[key][i][key2].asdict(),
                                    None,
                                )
                            else:
                                assert np.all(
                                    old_results[key][i][key2]
                                    == new_results[key][i][key2]
                                ), f"{key2} is not the same"
                else:
                    assert (
                        old_results[key] == new_results[key]
                    ), f"{key} is not the same"


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import matplotlib.pyplot as plt
import pytest
import sys

from openqaoa import QAOA, create_device, QAOABenchmark
from openqaoa.problems import QUBO
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS


def there_is_an_error(function, **kwargs):
    error = False
    try:
        function(**kwargs)
    except:
        error = True
    return error


class TestingBenchmark(unittest.TestCase):
    def test_simple_benchmark(self):
        "Test the simplest benchmark."

        qaoa = QAOA()
        qaoa.set_device(create_device(name="analytical_simulator", location="local"))
        qaoa.compile(QUBO.random_instance(5))

        benchmark = QAOABenchmark(qaoa)
        benchmark.run(n_points_axis=2**4, ranges=[(0, 2), (-2, 5)])

        assert benchmark.ranges == [
            (0, 2),
            (-2, 5),
        ], "The ranges saved are not correct."
        assert benchmark.ranges_reference == [
            (0, 2),
            (-2, 5),
        ], "The ranges saved for the reference are not correct."
        assert benchmark.values.shape == (
            16,
            16,
        ), "The shape of the values matrix is not correct."
        assert benchmark.values_reference.shape == (
            16,
            16,
        ), "The shape of the values matrix of the reference is not correct."
        assert benchmark.difference.shape == (
            16,
            16,
        ), "The shape of the difference matrix is not correct."
        assert all(
            benchmark.difference.flatten() == 0
        ), "The difference matrix should be all 0."
        assert benchmark.difference_mean == 0, "The difference mean should be 0."

    def test_type_qaoa_reference(self):
        "Test the type of the qaoa reference."

        qaoa_analytical = QAOA()
        qaoa_analytical.set_device(
            create_device(name="analytical_simulator", location="local")
        )
        qaoa_analytical.compile(QUBO.random_instance(5))

        qaoa_vectorized = QAOA()
        qaoa_vectorized.set_device(create_device(name="vectorized", location="local"))
        qaoa_vectorized.compile(QUBO.random_instance(5))

        # standard qaoa should have a analytical reference
        qaoa = QAOA()
        qaoa.set_device(create_device(name="vectorized", location="local"))
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)
        assert isinstance(
            benchmark.reference, QAOA
        ), "The qaoa reference should be a QAOA object."
        assert isinstance(
            benchmark.reference.backend, type(qaoa_analytical.backend)
        ), "The qaoa reference should have an analytical backend."

        # p>1 qaoa should have a vectorized reference
        qaoa = QAOA()
        qaoa.set_device(create_device(name="vectorized", location="local"))
        qaoa.set_circuit_properties(p=2)
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)
        assert isinstance(
            benchmark.reference, QAOA
        ), "The qaoa reference should be a QAOA object."
        assert isinstance(
            benchmark.reference.backend, type(qaoa_vectorized.backend)
        ), "The qaoa reference should have a vectorized backend."

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def __compare_values_benchmark(
        self,
        qaoa,
        n_points_axis,
        ranges,
        run_reference=False,
        verbose=True,
        plot_options={},
    ):
        "private function to compare the values of the benchmark with the values obtained by evaluating the circuit manually."

        benchmark = QAOABenchmark(qaoa)
        benchmark.run(
            n_points_axis=n_points_axis,
            ranges=ranges,
            run_reference=run_reference,
            verbose=verbose,
            plot_options=plot_options,
        )

        ranges_to_use = [r for r in ranges if len(r) == 2]
        values = np.zeros([n_points_axis for _ in range(len(ranges_to_use))])
        axes = [np.linspace(r[0], r[1], n_points_axis) for r in ranges_to_use]

        params = lambda x: [r[0] if len(r) == 1 else x.pop(0) for r in ranges]

        if len(ranges_to_use) == 1:
            for i in range(n_points_axis):
                values[i] = qaoa.evaluate_circuit(params([axes[0][i]]))["cost"]
        elif len(ranges_to_use) == 2:
            for i in range(n_points_axis):
                for j in range(n_points_axis):
                    values[j, i] = qaoa.evaluate_circuit(
                        params([axes[0][i], axes[1][j]])
                    )["cost"]
        elif len(ranges_to_use) == 3:
            for i in range(n_points_axis):
                for j in range(n_points_axis):
                    for k in range(n_points_axis):
                        values[k, j, i] = qaoa.evaluate_circuit(
                            params([axes[0][i], axes[1][j], axes[2][k]])
                        )["cost"]
        else:
            raise ValueError(
                "Bug in the tests: The number of ranges must be 1, 2 or 3."
            )

        assert (
            benchmark.ranges == ranges
        ), "The ranges saved are not correct. Inputs: n_points_axis={}, ranges={}".format(
            n_points_axis, ranges
        )
        assert (
            benchmark.values.shape == values.shape
        ), "The shape of the values matrix is not correct. Inputs: n_points_axis={}, ranges={}".format(
            n_points_axis, ranges
        )
        assert all(
            benchmark.values.flatten() == values.flatten()
        ), "The values matrix is not correct. Inputs: n_points_axis={}, ranges={}".format(
            n_points_axis, ranges
        )

        if run_reference:
            assert (
                benchmark.ranges_reference == ranges
            ), "The ranges saved for the reference are not correct. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )
            assert (
                benchmark.values_reference.shape == values.shape
            ), "The shape of the values matrix of the reference is not correct. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )
            assert (
                benchmark.difference.shape == values.shape
            ), "The difference matrix should have the same shape as the values matrix. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )
            assert isinstance(
                benchmark.difference_mean, float
            ), "The difference mean should be a float. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )
        else:
            assert (
                benchmark.values_reference is None
            ), "There should be no reference values, since run_reference=False. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )
            assert (
                benchmark.ranges_reference is None
            ), "There should be no reference ranges, since run_reference=False. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )

            error = False
            try:
                benchmark.difference
            except:
                error = True
            assert (
                error
            ), "There should be no difference matrix, since run_reference=False. An error should be raised. Inputs: n_points_axis={}, ranges={}, run_reference={}".format(
                n_points_axis, ranges, run_reference
            )

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_values_1D(self):
        "Test the values of the benchmark for 1D ranges."

        # standard case
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        self.__compare_values_benchmark(
            qaoa, n_points_axis=10, ranges=[(-24.78, 65.97), (356.467,)]
        )

        # extended parameters
        qaoa = QAOA()
        qaoa.set_circuit_properties(param_type="extended")
        qaoa.compile(
            QUBO(
                n=5,
                terms=[[0, 1], [1, 2], [2, 3], [3, 4], [4, 2], [4, 0]],
                weights=[1, 2, 1, 2, 1, 2],
            )
        )
        self.__compare_values_benchmark(
            qaoa,
            n_points_axis=10,
            ranges=[
                (-24.78,),
                (65.97,),
                (356.467,),
                (-24.78,),
                (65.97,),
                (0.03, 0.031),
                (356.467,),
                (2,),
                (-24.78,),
                (65.97,),
                (356.467,),
            ],
        )

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_values_2D(self):
        "Test the values of the benchmark for 2D ranges."

        # standard case
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        self.__compare_values_benchmark(
            qaoa,
            n_points_axis=4,
            ranges=[(-1.43, 2.78), (12.3, 25.68)],
            verbose=False,
            plot_options={"verbose": False},
        )

        # p=2
        qaoa = QAOA()
        qaoa.set_circuit_properties(p=2)
        qaoa.compile(QUBO.random_instance(5))
        self.__compare_values_benchmark(
            qaoa,
            n_points_axis=4,
            ranges=[(-1.43, 2.78), (12.3,), (0.5, 1.5), (1,)],
            run_reference=True,
        )

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_values_3D(self):
        "Test the values of the benchmark for 3D ranges."

        # standard case
        qaoa = QAOA()
        qaoa.set_circuit_properties(p=2)
        qaoa.compile(QUBO.random_instance(5))
        self.__compare_values_benchmark(
            qaoa,
            n_points_axis=4,
            ranges=[(-1.43, 2.78), (12.3, 25.68), (0.5, 1.5), (1,)],
            verbose=True,
        )

        # p=3
        qaoa = QAOA()
        qaoa.set_circuit_properties(p=3)
        qaoa.compile(QUBO.random_instance(5))
        self.__compare_values_benchmark(
            qaoa,
            n_points_axis=4,
            ranges=[(-1.43, 2.78), (245,), (0.5, 1.5), (1,), (0.5, 1.5), (19,)],
            run_reference=True,
        )

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_difference(self):
        "Test the property difference and difference_mean."

        if "qiskit.shot_simulator" not in SUPPORTED_LOCAL_SIMULATORS:
            self.skipTest(
                reason="The Qiskit Shot Simulator is currently not available. Please install the openqaoa-qiskit plugin."
            )

        # device shot-based
        qaoa = QAOA()
        qaoa.set_device(create_device(location="local", name="qiskit.shot_simulator"))
        qaoa.compile(QUBO.random_instance(5))

        benchmark = QAOABenchmark(qaoa)
        benchmark.run(n_points_axis=4, ranges=[(0, np.pi), (-5, 9)])

        assert (
            np.abs(benchmark.difference_mean) > 0
        ), "The difference mean should be non-zero."
        assert np.all(
            benchmark.difference == benchmark.values - benchmark.values_reference
        ), "The difference matrix should be the difference between the values and the reference values."
        assert benchmark.difference_mean == np.mean(
            benchmark.difference
        ), "The difference mean should be the mean of the difference matrix."

    def test_difference(self):
        "Test the property difference and difference_mean."

        # test raise errors if not run main or run reference
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        error = False
        try:
            benchmark.difference
        except:
            error = True
        assert (
            error
        ), "An error should be raised if the difference is called before running the benchmark."

        benchmark.run(
            n_points_axis=4,
            ranges=[(-np.pi, 23), (3, 4)],
            run_reference=True,
            run_main=False,
        )
        error = False
        try:
            benchmark.difference
        except:
            error = True
        assert (
            error
        ), "An error should be raised if the difference is called before running the benchmark."

        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        benchmark.run(
            n_points_axis=4,
            ranges=[(-np.pi, 23), (3, 4)],
            run_reference=False,
            run_main=True,
        )
        error = False
        try:
            benchmark.difference
        except:
            error = True
        assert (
            error
        ), "An error should be raised if the difference is called before running the benchmark."

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_input_checks(self):
        "Test the input assertions of the run method."

        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        assert there_is_an_error(
            benchmark.run,
            ranges=[(1, 2), (1, 3)],
        ), "An error should be raised when n_points_axis is not passed."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
        ), "An error should be raised when ranges is not passed."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4.5,
            ranges=[(1, 2), (1, 3)],
        ), "An error should be raised when n_points_axis is not an integer."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=56,
        ), "An error should be raised when ranges is not a list."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[
                (1, 2),
            ],
        ), "An error when len(ranges) != len(qaoa.circuit_parameters)."
        assert there_is_an_error(
            benchmark.run, n_points_axis=4, ranges=[(1, 2), (3,), (34,)]
        ), "An error when len(ranges) != len(qaoa.circuit_parameters)."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1, 2), "yes"],
        ), "An error should be raised when ranges is not a list of tuples."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1, 2), (1, 3, 4)],
        ), "An error should be raised when ranges is not a list of tuples of length 2."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1,), (1,)],
        ), "An error should be raised when ranges all ranges are tuples of length 1."
        assert there_is_an_error(
            benchmark.run, n_points_axis=4, ranges=[(1, 2), (1, 3)], run_main="fhn"
        ), "An error should be raised when run_main or run_reference are not booleans."
        assert there_is_an_error(
            benchmark.run, n_points_axis=4, ranges=[(1, 2), (1, 3)], run_reference="fhn"
        ), "An error should be raised when run_main or run_reference are not booleans."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1, 2), (1, 3)],
            plot="fhn",
            plot_reference=True,
        ), "An error should be raised when plot or plot_reference are not booleans."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1, 2), (1, 3)],
            plot=True,
            plot_reference="fhn",
        ), "An error should be raised when plot or plot_reference are not booleans."
        assert there_is_an_error(
            benchmark.run,
            n_points_axis=4,
            ranges=[(1, 2), (1, 2)],
            run_main=False,
            run_reference=False,
        ), "An error should be raised if nor run_main nor run_reference are True."
        assert there_is_an_error(
            benchmark.run, n_points_axis=4, ranges=[(1, 2), (1, 3)], plot_options="fhn"
        ), "An error should be raised when plot_options is not a dictionary."
        assert there_is_an_error(
            benchmark.run, n_points_axis=4, ranges=[(1, 2), (1, 3)], verbose="d"
        ), "An error should be raised when verbose is not a boolean."

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_plot(self):
        "Test the plot method."

        # test standard plot
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        # 2D
        benchmark.run(n_points_axis=4, ranges=[(0, np.pi), (-5, 9)])
        benchmark.plot()
        benchmark.plot(main=False, reference=True)
        benchmark.plot(main=False, difference=True)
        benchmark.plot(main=True, reference=True, difference=True)
        benchmark.plot(main=False, reference=True, difference=True)
        benchmark.plot(main=True, reference=False, difference=True)
        benchmark.plot(main=True, reference=True, difference=False)

        # 1D
        benchmark.run(n_points_axis=4, ranges=[(0, np.pi), (1,)])
        benchmark.plot()
        benchmark.plot(main=False, reference=True)
        benchmark.plot(main=False, difference=True)
        benchmark.plot(main=True, reference=True, difference=True)
        benchmark.plot(main=False, reference=True, difference=True)
        benchmark.plot(main=True, reference=False, difference=True)
        benchmark.plot(main=True, reference=True, difference=False)
        benchmark.plot(main=True, reference=True, difference=True, one_plot=True)
        benchmark.plot(main=False, reference=True, difference=True, one_plot=True)
        benchmark.plot(main=True, reference=False, difference=True, one_plot=True)
        benchmark.plot(main=True, reference=True, difference=False, one_plot=True)

        # 3D
        qaoa = QAOA()
        qaoa.set_circuit_properties(p=2)
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)
        benchmark.run(n_points_axis=4, ranges=[(0, np.pi), (-5, 9), (0, 1), (1,)])
        benchmark.plot()

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_plot_inputs(self):
        "Test the plot method with different inputs."

        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        # 2D
        benchmark.run(n_points_axis=4, ranges=[(-np.pi / 2, np.pi / 2), (1, 3)])
        benchmark.plot(labels=["a", "b"], plot_options={"cmap": "gray"})
        benchmark.plot(
            main=True,
            reference=True,
            difference=True,
            labels=["a", "b"],
            title=["a", "b", "c"],
            labels_legend="test",
            plot_options={"cmap": "gray"},
        )
        benchmark.plot(
            main=True, reference=True, difference=True, verbose=False, title="s"
        )
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        benchmark.plot(ax=ax)
        plt.show()
        plt.close(fig)

        # 1D
        benchmark.run(n_points_axis=4, ranges=[(-np.pi / 2, np.pi / 2), (1,)])
        benchmark.plot(labels=["a", "b"], plot_options={"linewidth": 2})
        benchmark.plot(
            main=True,
            reference=True,
            difference=True,
            labels=["a", "b"],
            title=["a", "b", "c"],
            plot_options={"linewidth": 2},
        )
        benchmark.plot(
            main=True,
            reference=True,
            difference=True,
            verbose=False,
            one_plot=True,
            title="s",
            labels_legend=["a", "b", "c"],
        )
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        benchmark.plot(ax=ax)
        plt.show()
        plt.close(fig)

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_plot_input_checks(self):
        "Test the plot method assertion inputs."

        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        benchmark.run(n_points_axis=4, ranges=[(0, np.pi), (-5, 9)])
        assert there_is_an_error(
            benchmark.plot, ax=""
        ), "An error should be raised when ax is not a matplotlib Axes"
        assert there_is_an_error(
            benchmark.plot, labels=""
        ), "An error should be raised when labels is not a list of two strings"
        assert there_is_an_error(
            benchmark.plot, labels=[""]
        ), "An error should be raised when labels is not a list of two strings"
        assert there_is_an_error(
            benchmark.plot, labels=["", "", ""]
        ), "An error should be raised when labels is not a list of two strings"
        assert there_is_an_error(
            benchmark.plot, main=""
        ), "An error should be raised when main is not a boolean"
        assert there_is_an_error(
            benchmark.plot, reference=""
        ), "An error should be raised when reference is not a boolean"
        assert there_is_an_error(
            benchmark.plot, difference=""
        ), "An error should be raised when difference is not a boolean"
        assert there_is_an_error(
            benchmark.plot, plot_options=""
        ), "An error should be raised when plot_options is not a dict"
        assert there_is_an_error(
            benchmark.plot, main=True, reference=True, difference=True, one_plot=True
        ), "An error should be raised when one_plot is True and more than one plot is requested when 2D"
        assert there_is_an_error(
            benchmark.plot, main=True, reference=True, difference=True, title=["a", "d"]
        ), "An error should be raised when one_plot is True and more than one plot is requested when 2D"

        # plot without values
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)

        assert there_is_an_error(
            benchmark.plot, main=True, reference=False
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=False, reference=True
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=False, reference=False, difference=True
        ), "An error should be raised when no values are available"

        benchmark.run(
            n_points_axis=4,
            ranges=[(0, np.pi), (1,)],
            run_main=True,
            run_reference=False,
        )
        assert there_is_an_error(
            benchmark.plot, main=True, reference=True
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=False, reference=True
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=False, reference=False, difference=True
        ), "An error should be raised when no values are available"

        benchmark = QAOABenchmark(qaoa)
        benchmark.run(
            n_points_axis=4,
            ranges=[(0, np.pi), (1,)],
            run_main=False,
            run_reference=True,
        )
        assert there_is_an_error(
            benchmark.plot, main=True, reference=True
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=True, reference=False
        ), "An error should be raised when no values are available"
        assert there_is_an_error(
            benchmark.plot, main=False, reference=False, difference=True
        ), "An error should be raised when no values are available"

    @pytest.mark.skipif(
        sys.platform == "darwin" or sys.platform.startswith("win"),
        reason="Test does not run on Mac and Windows currently",
    )
    def test_run_w_plots(self):
        qaoa = QAOA()
        qaoa.compile(QUBO.random_instance(5))
        benchmark = QAOABenchmark(qaoa)
        benchmark.run(
            n_points_axis=4,
            ranges=[(-np.pi / 2, np.pi / 2), (1, 3)],
            plot=True,
            run_reference=True,
            plot_difference=True,
        )



from .variational_params_converters import converter
from .variational_params_factory import (
    create_qaoa_variational_params,
    qaoa_variational_params_converter,
)

from .annealingparams import QAOAVariationalAnnealingParams
from .extendedparams import QAOAVariationalExtendedParams
from .fourierparams import (
    QAOAVariationalFourierParams,
    QAOAVariationalFourierExtendedParams,
    QAOAVariationalFourierWithBiasParams,
)
from .standardparams import (
    QAOAVariationalStandardParams,
    QAOAVariationalStandardWithBiasParams,
)



"""
Wavefunction simulator with methods focused on fast QAOA implementations.
Can easily be extended to do the full suite of operations in an ordinary simulator.
"""
from typing import Union, List, Tuple, Type, Optional
import numpy as np
from copy import copy
from scipy.sparse import csc_matrix, kron, diags
from scipy.sparse.linalg import expm

from .basebackend import QAOABaseBackendStatevector
from .gates_vectorized import VectorizedGateApplicator
from ..qaoa_components import QAOADescriptor, Hamiltonian
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ..utilities import generate_uuid, round_value


# Pauli gates
constI = csc_matrix(np.eye(2))
constX = csc_matrix(np.array([[0, 1], [1, 0]]))
constY = csc_matrix(np.array([[0, -1j], [1j, 0]]))
constZ = csc_matrix(np.array([[1, 0], [0, -1]]))

# Single qubit gates
constH = csc_matrix((1 / np.sqrt(2)) * np.array([[1, 1], [1, -1]]))
constS = csc_matrix(np.array([[1, 0], [0, 1j]]))
constT = csc_matrix(np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]]))

# Two-qubit gates
constCNOT = csc_matrix(
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
)
constCZ = csc_matrix(
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
)

# Projection (measurement) operators
P0 = csc_matrix(np.array([[1, 0], [0, 0]]))
P1 = csc_matrix(np.array([[0, 0], [0, 1]]))


# Parametrised rotations
def RX(theta: float) -> csc_matrix:
    return csc_matrix(expm(-1j * theta * constX / 2))


def RY(theta: float) -> csc_matrix:
    return csc_matrix(expm(-1j * theta * constY / 2))


def RZ(theta: float) -> csc_matrix:
    return csc_matrix(expm(-1j * theta * constZ / 2))


def CR_Z(theta: float) -> csc_matrix:
    return kron(P0, constI, format="csc") + kron(P1, RZ(theta), format="csc")


def ZZ(theta: float) -> csc_matrix:
    return diags([1, np.exp(-1j * theta), np.exp(-1j * theta), 1], 0, format="csc")


def CPHASE(theta: float) -> csc_matrix:
    return diags([1, 1, 1, np.exp(1j * theta)], 0, format="csc")


def _get_perm(n_qubits: int, qubits: list) -> list:
    """
    Helper function for several methods below.

    Parameters
    ----------
    n_qubits:
        total number of qubits in the register

    qubits:
        the qubits being permuted to the beginning of the register

    Returns
    -------
    perm:
        the permutation to apply to the wavefunction to bring the
        active qubits to the start of the register

    perminv:
        the inverse permutation, taking them back to where they belong
        from the beginning
    """

    # Get modified indices of qubits to coincide with indexing from the right
    qubits = [np.arange(n_qubits)[::-1][i] for i in qubits]

    # Permute the active qubits to the beginning of the register
    other_qubits = list(set(np.arange(n_qubits)) - set(qubits))
    perm = qubits + other_qubits
    perminv = list(np.argsort(perm))

    return perm, perminv


def _permute_qubits(wavefn: np.ndarray, perm: list) -> np.ndarray:
    """
    Reorganises the wavefunction components according
    to the specified qubit permutation.

    Parameters
    ----------
    wavefn:
        The current wavefunction of the register

    perm:
        The permutation according to which the
        register is to be reorganised.

    Returns
    -------
    wavefn:
        The permuted wavefunction.
    """

    # Permute according to specified perm
    wavefn = np.transpose(wavefn, perm)

    return wavefn


def _build_cost_hamiltonian(
    n_qubits: int, cost_hamiltonian: Type[Hamiltonian]
) -> np.array:
    """
    Builds the cost Hamiltonian as a vector, since it is diagonal.
    Output is an ndarray of shape [2]*n_qubits,
    for use in the `expectation_value` method.

    Parameters
    ----------
    n_qubits:
        number of qubits

    cost_hamiltonian:
        Hamiltonian object containing information about
        single/2-qubit terms and their weights.

    Returns
    -------
    ham_op:
        the Hamiltonian as a diagonal matrix reshaped
        to a [2]*n_qubits dimensional array

    """

    bias_qubits = cost_hamiltonian.qubits_singles
    biases = cost_hamiltonian.single_qubit_coeffs
    pairs = cost_hamiltonian.qubits_pairs
    coeffs = cost_hamiltonian.pair_qubit_coeffs

    terms = cost_hamiltonian.terms
    weights = cost_hamiltonian.coeffs

    # Check for non-classical terms
    cost_ham_pauli_str_lst = [term.pauli_str for term in terms]
    for term in cost_ham_pauli_str_lst:
        if str(term) != "Z" and str(term) != "ZZ":
            raise Exception(
                f"Currently, only classical cost Hamiltonians"
                "that consists of 'Z' and 'ZZ' terms are supported,"
                "but a '{term}' term was encountered."
            )

    ## ZZ operator on first two qubits, identity on all others to the right
    # Diagonal matrix (vector) for cost function
    iden_plus = np.ones(2 ** (n_qubits - 2), dtype=int)
    iden_minus = -1 * np.ones(2 ** (n_qubits - 1), dtype=int)
    ZZ_op = np.hstack((iden_plus, np.hstack((iden_minus, iden_plus))))
    ZZ_op.shape = [2] * n_qubits

    ## Z operator on first qubit, identity on all others to the right
    Z_op = np.hstack(
        (
            np.ones(2 ** (n_qubits - 1), dtype=int),
            -1 * np.ones(2 ** (n_qubits - 1), dtype=int),
        )
    )
    Z_op.shape = [2] * n_qubits

    ham_op = np.zeros([2] * n_qubits)
    for j in range(len(pairs)):
        perm, perminv = _get_perm(n_qubits, pairs[j].qubit_indices)
        ZZ_op = _permute_qubits(ZZ_op, perminv)

        ham_op += coeffs[j] * ZZ_op

        ZZ_op = _permute_qubits(ZZ_op, perm)

    for j in range(len(bias_qubits)):
        perm, perminv = _get_perm(n_qubits, [bias_qubits[j].qubit_indices])
        Z_op = _permute_qubits(Z_op, perminv)

        ham_op += biases[j] * Z_op

        Z_op = _permute_qubits(Z_op, perm)

    # add the constant term from the hamiltonian
    ham_op += cost_hamiltonian.constant

    return ham_op


class QAOAvectorizedBackendSimulator(QAOABaseBackendStatevector):
    r"""
    A simulator class for quantum circuits, oriented to QAOA,
    and more generally unitaries generated by Hamiltonians which
    consists of sums of Pauli strings. Works by translating the actions
    of single and two-Pauli rotation gates into permutations of wavefunction coefficients,
    obtained by slicing the (2, 2, ..., 2)-shaped wavefunction.

    Procedure:

        * Decompose rotation matrices into sum of identity and Pauli matrices with Euler's formula.
        * Compute the action of the Pauli matrices
            * Pauli X matrix : Flip coefficients with 0 at i-th qubit with
              coefficients with 1 at i-th qubit
            * Pauli Y matrx : Multiply 1j to everything. Flip coefficients with 0
              at i-th qubit with coefficients with 1 at i-th qubit, and
              multiply -1 to coefficients with 0 at i-th qubit.
            * Pauli Z matrix : multiply -1 to coefficients with 0 at i-th qubit.
        * Obtain final wavefunction by summing up
          :math:`\sin(\theta/2)* \textit{original wavefunction} - 1j*\cos(\theta/2)*\textit{processed wavefunction}`.

    Qubit labelling begins from the right, so that the right-most qubit has label 0,
    and the left-most has label n_qubits-1.

    Parameters
    ----------
    qaoa_descriptor: QAOADescriptor
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
    prepend_state: np.array
        The initial state of the circuit (before Hadamards).
        An array of shape :math:`(2^{n_qubits},)` or (2, 2, ..., 2).
        Defaults to ``[1,0,...,0]`` if ``None``.
    append_state: np.array
        A unitary matrix of shape :math:`(2^{self.n_qubits}, 2^{self.n_qubits})`,
        to be multiplied to the output state.
    init_hadamard: bool
        Whether to apply Hadamard gates to the beginning of the QAOA part of the circuit.
    cvar_alpha: float
        Conditional Value-at-Risk (CVaR) - a measure that takes into account
        only the tail of the probability distribution arising from the circut's
        count dictionary. Must be between 0 and 1.
        Check https://arxiv.org/abs/1907.04769 for further details.
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        prepend_state: Optional[Union[np.ndarray, List[complex]]],
        append_state: Optional[Union[np.ndarray, List[complex]]],
        init_hadamard: bool,
        cvar_alpha: float = 1,
    ):
        assert (
            cvar_alpha == 1
        ), "Please use the shot-based simulator for simulations with cvar_alpha < 1"

        QAOABaseBackendStatevector.__init__(
            self,
            qaoa_descriptor,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )

        # Build the Hamiltonian operator as an array
        self.ham_op = _build_cost_hamiltonian(self.n_qubits, self.cost_hamiltonian)

        if self.n_qubits > 0:
            self.wavefn = np.zeros((2**self.n_qubits,), dtype=complex)
            self.wavefn[0] = 1
            self.wavefn = self.wavefn.reshape([2] * self.n_qubits)
        else:
            self.wavefn = []

        # Handle prepend state
        if self.prepend_state is not None:
            if isinstance(self.prepend_state, np.ndarray):
                if np.shape(self.prepend_state) == np.shape(self.wavefn):
                    self.wavefn = self.prepend_state
                elif np.shape(self.prepend_state) == (2**self.n_qubits,):
                    self.wavefn = self.prepend_state.reshape([2] * self.n_qubits)
                else:
                    raise ValueError(
                        "Error : Unsupported prepend_state specified."
                        "Not of shape (2**n,) or (2, 2, ..., 2))."
                    )

            else:
                raise ValueError(
                    "Error : Unsupported prepend_state specified. Not an ndarray."
                )

        # Handle append state
        if self.append_state is not None:
            if isinstance(self.append_state, np.ndarray) and np.shape(
                self.append_state
            ) == (2**self.n_qubits, 2**self.n_qubits):
                # check unitarity of append_state matrix
                if not np.allclose(
                    np.eye(2**self.n_qubits),
                    self.append_state.dot(self.append_state.conj().T),
                ):
                    raise ValueError("append_state is not a unitary matrix")

            else:
                raise ValueError(
                    "Unsupported append_state specified (Not an ndarray,"
                    "or not of shape (2**n, 2**n)."
                )

        # Handle init_hadamard
        if self.init_hadamard:
            for i in range(self.n_qubits):
                self.apply_hadamard(i)

        # store the initialisation part of wavefunction
        self.wavefn_init = copy(self.wavefn)

    # Apply gate methods
    def apply_x(self, qubit_1: int):
        r"""
        Applies the X gate on ``qubit_1`` in a vectorized way.

        Parameters
        ----------
        qubit_1:
            Qubit index to apply gate.

        Returns
        -------
            None
        """

        rotation_angle = np.pi
        C = np.cos(rotation_angle / 2)
        S = -1j * np.sin(rotation_angle / 2)
        wfn = (C * self.wavefn) + (
            S * np.flip(self.wavefn, self.n_qubits - qubit_1 - 1)
        )

        self.wavefn = wfn

    def apply_rx(self, qubit_1: int, rotation_angle: float):
        r"""
        Applies the RX($\theta$ = ``rotation_angle``) gate on ``qubit_1`` in a vectorized way.
        
        **Definition of RX($\theta$):**

        .. math::

            RX(\theta) = \exp\left(-i \frac{\theta}{2} X\right) =
            \begin{pmatrix}
                \cos{\frac{\theta}{2}}   & -i\sin{\frac{\theta}{2}} \\
                -i\sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}}
            \end{pmatrix}
            
        Parameters
        ----------
        qubit_1:
            Qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        C = np.cos(rotation_angle / 2)
        S = -1j * np.sin(rotation_angle / 2)
        wfn = (C * self.wavefn) + (
            S * np.flip(self.wavefn, self.n_qubits - qubit_1 - 1)
        )

        self.wavefn = wfn

    def apply_ry(self, qubit_1: int, rotation_angle: float):
        r"""
        Applies the RY($\theta$ = ``rotation_angle``) gate on ``qubit_1`` in a vectorized way.
        
        **Definition of RY($\theta$):**

        .. math::

            RY(\theta) = \exp\left(-i \frac{\theta}{2} Y\right) =
            \begin{pmatrix}
                \cos{\frac{\theta}{2}} & -\sin{\frac{\theta}{2}} \\
                \sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}}
            \end{pmatrix}

        Parameters
        ----------
        qubit_1:
            Qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        wfn = copy(self.wavefn)

        # multiply slices with i/-i
        slc_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_0] *= -1j
        wfn[slc_1] *= 1j

        C = np.cos(rotation_angle / 2)
        S = 1j * np.sin(rotation_angle / 2)

        self.wavefn = (C * self.wavefn) + (
            S * np.flip(wfn, self.n_qubits - qubit_1 - 1)
        )

    def apply_rz(self, qubit_1: int, rotation_angle: float):
        r"""
        Applies the RZ($\theta$ = ``rotation_angle``) gate on ``qubit_1`` in a vectorized way.
        
        **Definition of RZ($\theta$):**

        .. math::

            RZ(\theta) = \exp\left(-i \frac{\theta}{2} Z\right) =
                \begin{pmatrix}
                    e^{-i\frac{\theta}{2}} & 0 \\
                    0 & e^{i \frac{\theta}{2}}
                \end{pmatrix}
            
        Parameters
        ----------
        qubit_1:
            Qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        slc_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )

        self.wavefn[slc_0] *= np.exp(-1j * rotation_angle / 2)
        self.wavefn[slc_1] *= np.exp(1j * rotation_angle / 2)

    def apply_rxx(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        r"""
        Applies the RXX($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.
        
        **Definition of RXX($\theta$):**

        .. math::

            R_{XX}(\theta) = \exp\left(-i \frac{\theta}{2} X{\otimes}X\right) =
                \begin{pmatrix}
                    \cos\left(\frac{\theta}{2}\right)   & 0           & 0           & -i\sin\left(\frac{\theta}{2}\right) \\
                    0           & \cos\left(\frac{\theta}{2}\right)   & -i\sin\left(\frac{\theta}{2}\right) & 0 \\
                    0           & -i\sin\left(\frac{\theta}{2}\right) & \cos\left(\frac{\theta}{2}\right)   & 0 \\
                    -i\sin\left(\frac{\theta}{2}\right) & 0           & 0           & \cos\left(\frac{\theta}{2}\right)
                \end{pmatrix}
            

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.
            
        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        # SH TODO : investigate if slicing 01 and 10 coefficients and swapping once them is faster than flipping twice
        C = np.cos(rotation_angle / 2)
        S = -1j * np.sin(rotation_angle / 2)

        wfn = (C * self.wavefn) + (
            S
            * np.flip(
                np.flip(self.wavefn, self.n_qubits - qubit_1 - 1),
                self.n_qubits - qubit_2 - 1,
            )
        )
        self.wavefn = wfn

    def apply_ryy(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        r"""
        Applies the RYY($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.
        
        **Definition of RYY($\theta$):**

        .. math::

            R_{YY}(\theta) = \exp\left(-i \frac{\theta}{2} X{\otimes}X\right) =
                \begin{pmatrix}
                    \cos\left(\frac{\theta}{2}\right)   & 0           & 0           & -i\sin\left(\frac{\theta}{2}\right) \\
                    0           & \cos\left(\frac{\theta}{2}\right)   & -i\sin\left(\frac{\theta}{2}\right) & 0 \\
                    0           & -i\sin\left(\frac{\theta}{2}\right) & \cos\left(\frac{\theta}{2}\right)   & 0 \\
                    -i\sin\left(\frac{\theta}{2}\right) & 0           & 0           & \cos\left(\frac{\theta}{2}\right)
                \end{pmatrix}
            

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.
            
        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        wfn = copy(self.wavefn)

        slc_q1_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q1_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q2_0 = tuple(
            0 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q2_1 = tuple(
            1 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )

        wfn[slc_q1_0] *= 1j
        wfn[slc_q1_1] *= -1j

        wfn[slc_q2_0] *= -1j
        wfn[slc_q2_1] *= 1j

        C = np.cos(rotation_angle / 2)
        S = 1j * np.sin(rotation_angle / 2)

        self.wavefn = (C * self.wavefn) + (
            S
            * np.flip(
                np.flip(wfn, self.n_qubits - qubit_1 - 1), self.n_qubits - qubit_2 - 1
            )
        )

    def apply_rzz(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        r"""
        Applies the RZZ($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.
        
        **Definition of RZZ($\theta$):**

        .. math::

            RZZ(\theta) = \exp\left(-i \frac{\theta}{2} Z{\otimes}Z\right) =
                \begin{pmatrix}
                    e^{-i \frac{\theta}{2}} & 0 & 0 & 0 \\
                    0 & e^{i \frac{\theta}{2}} & 0 & 0 \\
                    0 & 0 & e^{i \frac{\theta}{2}} & 0 \\
                    0 & 0 & 0 & e^{-i \frac{\theta}{2}}
                \end{pmatrix}
            
        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.
            
        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        """
        # Note : one can also slice the 01 and 10 elements:
        slc_pair01 = tuple(1 if i == self.n_qubits - qubit_1 - 1
                           else 0 if i == self.n_qubits - qubit_2 - 1
                           else slice(None) for i in range(self.n_qubits))
        slc_pair10 = tuple(1 if i == self.n_qubits - qubit_2 - 1
                           else 0 if i == self.n_qubits - qubit_1 - 1
                           else slice(None) for i in range(self.n_qubits))
        """

        slc_pair00 = tuple(
            0
            if i
            in [
                self.n_qubits - qubit_1 - 1,
                self.n_qubits - qubit_2 - 1,
            ]
            else slice(None)
            for i in range(self.n_qubits)
        )
        slc_pair11 = tuple(
            1
            if i
            in [
                self.n_qubits - qubit_1 - 1,
                self.n_qubits - qubit_2 - 1,
            ]
            else slice(None)
            for i in range(self.n_qubits)
        )

        self.wavefn[slc_pair00] *= np.exp(-1j * rotation_angle)
        self.wavefn[slc_pair11] *= np.exp(-1j * rotation_angle)
        self.wavefn *= np.exp(1j * rotation_angle / 2)

    def apply_rxy(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        """
        Applies the RXY($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.

        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        wfn = copy(self.wavefn)

        # Action of Y part
        slc_q2_0 = tuple(
            0 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q2_1 = tuple(
            1 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_q2_0] *= -1j
        wfn[slc_q2_1] *= 1j

        C = np.cos(rotation_angle / 2)
        S = 1j * np.sin(rotation_angle / 2)
        self.wavefn = (C * self.wavefn) + (
            S
            * np.flip(
                np.flip(wfn, self.n_qubits - qubit_1 - 1), self.n_qubits - qubit_2 - 1
            )
        )

    def apply_rzx(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        """
        Applies the RZX($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.

        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        wfn = copy(self.wavefn)

        # Action of Z part
        slc_q2_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q2_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_q2_0] *= 1
        wfn[slc_q2_1] *= -1

        C = np.cos(rotation_angle / 2)
        S = -1j * np.sin(rotation_angle / 2)
        self.wavefn = (C * self.wavefn) + (
            S * np.flip(wfn, self.n_qubits - qubit_2 - 1)
        )

    def apply_ryz(self, qubit_1: int, qubit_2: int, rotation_angle: float):
        """
        Applies the RYZ($\theta$ = ``rotation_angle``) gate on ``qubit_1`` and ``qubit_2`` in a vectorized way.

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.

        qubit_2:
            Second qubit index to apply gate.

        rotation_angle:
            Angle to be rotated.

        Returns
        -------
            None
        """

        wfn = copy(self.wavefn)

        # Action of Y part
        slc_q1_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q1_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_q1_0] *= -1j
        wfn[slc_q1_1] *= 1j

        # Action of Z part
        slc_q2_0 = tuple(
            0 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_q2_1 = tuple(
            1 if i == self.n_qubits - qubit_2 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_q2_0] *= 1
        wfn[slc_q2_1] *= -1

        C = np.cos(rotation_angle / 2)
        S = 1j * np.sin(rotation_angle / 2)
        self.wavefn = (C * self.wavefn) + (
            S * np.flip(wfn, self.n_qubits - qubit_1 - 1)
        )

    def apply_hadamard(self, qubit_1: int):
        """
        Applies the Hadamard gate on ``qubit_1`` in a vectorized way. Only used when ``init_hadamard`` is true.

        Parameters
        ----------
        qubit_1:
            First qubit index to apply gate.

        Returns
        -------
            None
        """
        # TODO : Combine init_hadamard and prepend_state into one.
        # vectorized hadamard gate, for when init_hadamard = True
        wfn = copy(self.wavefn)

        slc_0 = tuple(
            0 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        slc_1 = tuple(
            1 if i == self.n_qubits - qubit_1 - 1 else slice(None)
            for i in range(self.n_qubits)
        )
        wfn[slc_1] *= -1
        wfn[slc_0] += self.wavefn[slc_1]
        wfn[slc_1] += self.wavefn[slc_0]

        self.wavefn = wfn / np.sqrt(2)

    def qaoa_circuit(self, params: Type[QAOAVariationalBaseParams]):
        """
        Executes the entire QAOA circuit, with angles specified within ``params``.
        Steps:
        1) Creates a (2,...,2) dimensional matrix that represents a 2**n dimensional wavefunction
        2) Modify it according to the ``prepend_state`` option.
        3) Modify it according to ``init_hadamard`` option.
        4) Modify it according to list of gates in ``params``.
        5) Modify it accoding to ``append_state`` option.

        Parameters
        ----------
        params:
            ``QAOAVariationalBaseParams`` object that contains rotation angles and gates to be applied.

        Returns
        -------
            None
        """
        gates_applicator = VectorizedGateApplicator()

        # generate a job id for the wavefunction evaluation
        self.job_id = generate_uuid()

        # reset the wavefunction back to its initialisation state
        self.reset_circuit()

        # Assign angles and apply gates
        self.assign_angles(params)

        low_level_gate_list = []
        for each_gate in self.abstract_circuit:
            low_level_gate_list.extend(each_gate.decomposition("trivial"))

        for each_tuple in low_level_gate_list:
            gate = each_tuple[0](gates_applicator, *each_tuple[1])
            gate.apply_gate(self)

        # Handle append state
        if self.append_state is not None:
            # Flatten (2,...,2) shaped wfn into a 2**n-dim column vector before multiplying with unitary matrix, ...
            self.wavefn = np.matmul(self.append_state, self.wavefn.flatten())
            # then re-shape it back to (2,...,2)
            self.wavefn = self.wavefn.reshape([2] * self.n_qubits)

    def wavefunction(self, params: Type[QAOAVariationalBaseParams] = None) -> list:
        """
        Get the wavefunction of the state produced by the parametric circuit.

        Parameters
        ----------
        params:
            The QAOA parameters - an object of one of the parameter classes, containing
            hyperparameters and variable parameters.

        Returns
        -------
        wf:
            A list of the wavefunction amplitudes.
        """

        self.qaoa_circuit(params)

        self.wavefn.shape = 2**self.n_qubits

        self.measurement_outcomes = self.wavefn.flatten()

        # Make format same as ProjectQ
        wf = [(component) for component in self.wavefn]

        return wf

    @round_value
    def expectation(self, params: Type[QAOAVariationalBaseParams]) -> float:
        """
        Call the execute function on the circuit to compute the
        expectation value of the Quantum Circuit w.r.t cost operator

        Returns
        -------
        exp_val:
            The expectation value of the cost function wrt the state generated by the circuit.
        """

        self.qaoa_circuit(params)

        # Reshape wavefunction
        wavefn_ = self.wavefn

        self.measurement_outcomes = wavefn_.flatten()

        # Compute the expectation value and its standard deviation
        ham_wf = self.ham_op * wavefn_
        exp_val = np.real(np.vdot(wavefn_, ham_wf))

        out = exp_val

        return out

    @round_value
    def expectation_w_uncertainty(
        self, params: Type[QAOAVariationalBaseParams]
    ) -> Tuple[float, float]:
        """
        Call the execute function on the circuit to compute the
        expectation value of the ``QuantumCircuit`` w.r.t cost operator
        along with its uncertainty

        Returns
        -------
        exp_val:
            The expectation value of the cost function wrt the state generated by the circuit.
        std_dev:
            The standard deviation of the cost function wrt the state generated by the circuit.
        """

        self.qaoa_circuit(params)

        # Reshape wavefunction
        wavefn_ = self.wavefn

        self.measurement_outcomes = wavefn_.flatten()

        # Compute the expectation value and its standard deviation
        ham_wf = self.ham_op * wavefn_
        exp_val = np.real(np.vdot(wavefn_, ham_wf))

        exp_val_sq = np.real(np.vdot(ham_wf, ham_wf))
        std_dev = (exp_val_sq - exp_val**2) ** 0.5
        out = exp_val, std_dev

        return out

    def reset_circuit(self):
        """
        Reset the circuit by resetting the wavefunction.
        """
        self.wavefn = copy(self.wavefn_init)

    def circuit_to_qasm(self):
        """
        A method to convert the entire QAOA ``QuantumCircuit`` object into
        a OpenQASM string
        """
        raise NotImplementedError()



import unittest

from openqaoa.backends import DeviceLocal
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS


class TestingDeviceLocal(unittest.TestCase):

    """
    This tests check that the Device Object created for local devices have the
    appropriate behaviour.
    """

    def test_supported_device_names(self):
        for each_device_name in SUPPORTED_LOCAL_SIMULATORS:
            device_obj = DeviceLocal(each_device_name)

            self.assertEqual(device_obj.check_connection(), True)

    def test_unsupported_device_names(self):
        device_obj = DeviceLocal("unsupported_device")

        self.assertEqual(device_obj.check_connection(), False)


if __name__ == "__main__":
    unittest.main()



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import SK


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestSK(unittest.TestCase):
    """Tests for the SK class"""

    def test_sk_terms_weights_constant(self):
        """Test that SK creates a correct QUBO from the provided graph"""

        n_nodes = 5
        rng = np.random.default_rng(1234)
        G = nx.Graph()
        G.add_weighted_edges_from(
            [
                [i, j, round(2 * rng.random() - 1)]
                for i in range(n_nodes)
                for j in range(i + 1, n_nodes)
            ]
        )
        gr_edges = [[0, 1], [0, 3], [1, 3], [1, 4], [2, 4], [0], [1], [2], [3], [4]]
        gr_weights = [-1.0, -1.0, 1.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        sk_prob_qubo = SK(G).qubo
        self.assertTrue(terms_list_equality(gr_edges, sk_prob_qubo.terms))
        self.assertEqual(gr_weights, sk_prob_qubo.weights)
        self.assertEqual(0.0, sk_prob_qubo.constant)

    def test_sk_random_problem(self):
        """Test SK random instance method"""

        seed = 123
        n_nodes = 5
        mu = 0
        sigma = 1
        rng = np.random.default_rng(seed)
        G = nx.Graph()
        G.add_weighted_edges_from(
            [
                [i, j, round(rng.normal(loc=mu, scale=sigma), 3)]
                for i in range(n_nodes)
                for j in range(i + 1, n_nodes)
            ]
        )
        sk_manual_prob = SK(G).qubo

        sk_random_prob = SK.random_instance(n_nodes=n_nodes, seed=seed).qubo

        self.assertTrue(terms_list_equality(sk_manual_prob.terms, sk_random_prob.terms))
        self.assertEqual(sk_manual_prob.weights, sk_random_prob.weights)
        self.assertEqual(sk_manual_prob.constant, sk_random_prob.constant)

    def test_sk_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {"node1": 1, "node2": 2}, np.array([1, 2])]

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                SK(G=each_graph)
            self.assertEqual(
                "Input problem graph must be a networkx Graph.", str(e.exception)
            )

    def test_sk_classical_sol(self):
        """Test the SK random instance method classical solution"""

        seed = 1234
        sk_sol = SK.random_instance(n_nodes=2, seed=seed).classical_solution()

        sol = {"x_0": 1, "x_1": 0}

        self.assertEqual(sk_sol, sol)

    def test_sk_plot(self):
        """Test SK random instance method"""
        import matplotlib.pyplot as plt

        seed = 1234
        sk_random_prob = SK.random_instance(n_nodes=10, seed=seed)
        sol = {
            "x_0": 1.0,
            "x_1": 1.0,
            "x_2": 0,
            "x_3": 1.0,
            "x_4": 0,
            "x_5": 1.0,
            "x_6": 0,
            "x_7": 0,
            "x_8": 0,
            "x_9": 1.0,
        }

        fig, ax = plt.subplots()
        sk_random_prob.plot_solution(sol, ax=ax)
        self.assertTrue(isinstance(ax, plt.Axes))


if __name__ == "__main__":
    unittest.main()



import networkx as nx
from docplex.mp.model import Model
from typing import Union
import matplotlib.pyplot as plt
import numpy as np

from ..utilities import check_kwargs
from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class MIS(Problem):
    """
    Creates an instance of the maximal independent set (MIS) problem.
    https://en.wikipedia.org/wiki/Maximal_independent_set

    Parameters
    ----------
    G: networkx.Graph
        Networkx graph of the problem
    penalty: float
        Penalty for the edge constraints.

    Returns
    -------
        An instance of the MIS problem.
    """

    __name__ = "maximal_independent_set"

    def __init__(self, G, penalty: Union[int, float] = 2):
        self.G = G
        self.penalty = penalty
        # self.cplex_model = self.docplex_model()

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the maximal independent set problem, whose graph is
        random following the Erdos-Renyi model.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
            A random instance of the maximal independent set problem.
        """
        n_nodes, edge_probability = check_kwargs(
            ["n_nodes", "edge_probability"], [None, None], **kwargs
        )
        seed = kwargs.get("seed", None)

        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )
        return MIS(G)

    @property
    def docplex_model(self):
        mdl = Model("MIS")
        num_vertices = self.G.number_of_nodes()

        x = {i: mdl.binary_var(name=f"x_{i}") for i in range(num_vertices)}

        mdl.minimize(
            -mdl.sum(x) + self.penalty * mdl.sum(x[i] * x[j] for (i, j) in self.G.edges)
        )
        return mdl

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        qubo_docplex = FromDocplex2IsingModel(cplex_model).ising_model

        return QUBO(
            self.G.number_of_nodes(),
            qubo_docplex.terms + [[]],
            qubo_docplex.weights + [qubo_docplex.constant],
            self.problem_instance,
        )

    def classical_solution(self, string: bool = False):
        """
        Return the classical solution of the maximal independent set problem
        Parameters
        ----------
        string : bool, optional
            If the solution is returned as a string. The default is False.
        Raises
        ------
        ValueError
            A flag if docplex does not find a valid solution.
        Returns
        -------
        solution : Union[str, dict]
            The classical solution of the specific problem as a string or a dict.
        """
        cplex_model = self.docplex_model
        docplex_sol = cplex_model.solve()

        if docplex_sol is None:
            raise ValueError(
                f"solution not found: {self.cplex_model.solve_details.status}"
            )

        if string:
            solution = ""
        else:
            solution = {}
        for var in cplex_model.iter_binary_vars():
            if string:
                solution += str(int(round(docplex_sol.get_value(var), 1)))
            else:
                solution[var.name] = int(round(docplex_sol.get_value(var), 1))
        return solution

    def plot_solution(self, solution: Union[dict, str], ax=None):
        """
        A visualization method for the maximal independet set solution.
        Parameters
        ----------
        solution : Union[dict, str]
            The solution of the specific bin packing problem as a string or dictionary.
        ax : matplotlib axes, optional
            The default is None.
        Returns
        -------
        fig : matplotlib.pyplot.Figure()
            The graph visualization of the solution.
        """
        if isinstance(solution, str):
            sol = self.solution.copy()
            for n, var in enumerate(self.docplex_model().iter_binary_vars()):
                sol[var.name] = int(solution[n])
            solution = sol
        colors = ["#5EB1EB", "#F29D55"]
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        pos = nx.circular_layout(self.G)
        num_vertices = self.G.number_of_nodes()
        nx.draw(
            self.G,
            pos=pos,
            edgecolors="black",
            labels={i: str(i) for i in range(num_vertices)},
            node_color=[colors[solution[f"x_{i}"]] for i in range(num_vertices)],
            ax=ax,
            edge_color="#0B2340",
        )
        ax.plot(
            [],
            [],
            marker="o",
            markersize=10,
            label="MIS",
            linewidth=0,
            color=colors[1],
            markeredgecolor="black",
        )
        ax.legend(loc="upper center", bbox_to_anchor=[0.5, 1.1])
        return fig



import random
import numpy as np
import json

from .basebackend import VQABaseBackend

from ..qaoa_components.ansatz_constructor.gates import X

from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)

from ..qaoa_components import Hamiltonian

from ..utilities import (
    exp_val_pair,
    exp_val_single,
    negate_counts_dictionary,
    calculate_calibration_factors,
    round_value,
)


class BaseWrapper(VQABaseBackend):
    """
    This is the Abstract Base Class over which other classes will be built.
    It is designed to take the backend object and wrap around it in order
    to override the necessary methods.

    Parameters
    ----------
    backend: `VQABaseBackend`
        The backend object to be wrapped.
    """

    def __init__(self, backend):
        self.backend = backend

    def __getattr__(self, name):
        return getattr(self.backend, name)

    def expectation(self, *args, **kwargs):
        return self.backend.expectation(*args, **kwargs)

    def expectation_w_uncertainty(self, *args, **kwargs):
        return self.backend.expectation_w_uncertainty(*args, **kwargs)

    def exact_solution(self, *args, **kwargs):
        return self.backend.exact_solution(*args, **kwargs)


class SPAMTwirlingWrapper(BaseWrapper):
    """
    This class inherits from the BaseWrapper and needs to be backend
    agnostic to QAOA implementations on different devices and their
    respective SDKs. It implements State Preparation and Measurement (SPAM)
    Twirling which is a simple error mitigation technique used to remove any
    preferred direction due to readout noise.

    Parameters
    ----------
    n_batches: `int`
        Number of batches.
    calibration_data_location: `str`
        The location of the calibration data file.
    """

    def __init__(self, backend, n_batches, calibration_data_location):
        super().__init__(backend)
        self.n_batches = n_batches
        self.calibration_data_location = calibration_data_location

        with open(self.calibration_data_location, "r") as f:
            calibration_data = json.load(f)

        calibration_measurements = calibration_data["results"]["measurement_outcomes"]
        calibration_registers = calibration_data["register"]

        assert (
            len(calibration_registers) >= self.backend.n_qubits
        ), "Problem requires more qubits than present in the calibration data."
        "Make sure that the calibration data file is for the correct device."

        qubit_mapping = self.backend.initial_qubit_mapping

        self.calibration_factors = calculate_calibration_factors(
            self.backend.qaoa_descriptor.cost_hamiltonian,
            calibration_measurements,
            calibration_registers,
            qubit_mapping,
        )

    def get_counts(
        self, params: QAOAVariationalBaseParams, n_shots=None, seed=None
    ) -> dict:
        """
        This method overrides the get_counts method of the backend object
        to obtain the measurement outcomes under the bit flip averaging
        technique, which is applying an X gate just before the measurement
        and negating the classical outcomes for a set of qubits selected at
        random. Every such set we call a batch, and the total number of shots
        is distributed amongst the number of batches. The procedure is
        as follows: per batch, generate a random integer, obtain its
        binary representation and use the positions of 1s in the bitstring to
        mark the qubits to which an X gate will be applied. Then, create an
        append state which is a set of all these X gates on the chosen qubits.
        Run the modified circuit and obtain the measurement outcomes using the
        original backend get_counts method. Negate the counts classically for
        the selected qubits. Lastly, combine the counts from all batches.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes,
            containing variable parameters.
        n_shots: `int`
            The number of shots to be used for the measurement. If None,
            the backend default.
        seed: `int`
            The seed controlling the random number generator, which selects
            the set of qubits to be negated at every batch. Default is None.

        Returns
        -------
        dict:
            The dictionary containing measurement outcomes (counts) of a given
            QAOA circit under the bit-flip-averaging (BFA) technique.
        """
        if seed is not None:
            random.seed(seed)
        # list of integers whose binary representation signifies which qubits to be flipped at every batch
        s_list = []
        for _ in range(0, self.n_batches):
            s_list.append(random.getrandbits(self.backend.n_qubits))

        n_shots = self.backend.n_shots if n_shots == None else n_shots
        n_shots_batch = int(n_shots / self.n_batches)

        counts = {}

        for batch in range(0, self.n_batches):
            s = s_list[batch]
            s_binary = format(s, "b").zfill(self.backend.n_qubits)  # convert to binary
            arr = np.fromiter(s_binary, dtype=int)
            negated_qubits = np.where(arr == 1)[0]  # where the syndrome has a 1

            circuit_to_append = self.backend.gate_applicator.create_quantum_circuit(
                self.backend.n_qubits
            )

            for negated_qubit in negated_qubits:
                negated_qubit = negated_qubit.item()
                negation_gate = X(self.backend.gate_applicator, negated_qubit)
                circuit_to_append = self.backend.gate_applicator.apply_gate(
                    negation_gate, negated_qubit, circuit_to_append
                )

            self.backend.append_state = circuit_to_append

            counts_batch = self.backend.get_counts(
                params, n_shots_batch
            )  # calls the get_counts method of the specific backend

            negated_counts = negate_counts_dictionary(
                counts_dictionary=counts_batch, s=s
            )

            # Add to the final counts dict
            for key in negated_counts:
                if key in counts:
                    counts[key] = counts[key] + negated_counts[key]
                else:
                    counts.update([(key, negated_counts[key])])

        self.measurement_outcomes = counts
        self.backend.append_state = []  # reset

        return counts

    @round_value
    def expectation_value_spam_twirled(
        self, counts: dict, hamiltonian: Hamiltonian, calibration_factors: dict
    ) -> float:
        """
        This method computes the expectation value of a Hamiltonian termwise
        while correcting for measurement errors by dividing by the
        hardware-specific calibration factors.

        Parameters
        ----------
        counts: `dict`
            A dictionary of the measurement outcomes of a given QAOA circuit.
        hamiltonian: `Hamiltonian`
            Hamiltonian object containing the problem statement.
        calibration_factors: `dict`
            Dictionary containing the relevant calibration factors computed
            given calibration data for the specific hardware.

        Returns
        -------
        float:
            The error mitigated expectation value of cost operator wrt a
            measurement outcomes dictionary (counts).
        """

        terms = [term.qubit_indices for term in hamiltonian.terms]
        hamiltonian_as_dict = hamiltonian.hamiltonian_dict()

        energy = 0

        # Compute expectation values and correlations of terms present in the Hamiltonian
        for term in hamiltonian.terms:
            # If bias term compute expectation value
            if len(term) == 1:
                i = term.qubit_indices[0]
                exp_vals_z = exp_val_single(i, counts)
                exp_vals_z /= calibration_factors[(i,)]

                energy += exp_vals_z * hamiltonian_as_dict[(i,)]

            # If two-body term compute correlation
            elif len(term) == 2:
                i, j = term.qubit_indices
                exp_vals_zz = exp_val_pair((i, j), counts)
                exp_vals_zz /= calibration_factors[(i, j)]

                energy += exp_vals_zz * hamiltonian_as_dict[(i, j)]

            # If constant term, ignore
            if len(term) == 0:
                continue

        energy += hamiltonian.constant

        return energy

    def expectation(self, params: QAOAVariationalBaseParams, n_shots=None) -> float:
        """
        This method overrrides the one from the basebackend to allow for
        correcting the expectation values of each term in the Hamiltonian
        before providing the energy to the optimizer. It does this by first
        obtaining the measurement outcomes through the modified get_counts
        method and then calculating the cost with the wrapper's own
        expectation_value_spam_twirled method.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes,
            containing variable parameters.
        n_shots: `int`
            The number of shots to be used for the measurement. If None,
            the backend default.

        Returns
        -------
        float:
            The error mitigated expectation value of cost operator wrt to
            quantum state produced by QAOA circuit
        """
        counts = self.get_counts(params, n_shots)
        cost = self.expectation_value_spam_twirled(
            counts,
            self.backend.qaoa_descriptor.cost_hamiltonian,
            self.calibration_factors,
        )

        return cost



import unittest
from docplex.mp.model import Model
import networkx as nx

from openqaoa.problems import QUBO, MaximumCut, FromDocplex2IsingModel


class TestDocplex2IsingClass(unittest.TestCase):

    """
    Test the converter from docplex models

    """

    def test_qubo(self):
        """
        Test the QUBO class is generated from the function FromDocplex2IsingModel

        """
        # Creating a basic docplex model
        mdl = Model("Test")  # Docplex model
        num_z = 5  # Number of variables
        z = mdl.binary_var_list(num_z, name="Z")  # docplex variables
        objective = mdl.sum(z) - 2 * z[0] + z[3] * z[4] + 5  # objective function

        mdl.minimize(objective)  # Optimization

        ising_problem = FromDocplex2IsingModel(
            mdl
        ).ising_model  # Ising model of the Docplex model

        self.assertIsInstance(ising_problem, QUBO)

    def test_slack_penalization(self):
        """
        Test the equality and inequality constraints are encoded in the QUBO
        model using slack variables for the inequality constraints approach.

        """
        weights = [
            18.25,
            -4.0,
            -8.0,
            -4.0,
            -6.0,
            -12.0,
            -6.0,
            4.0,
            2.0,
            4.0,
            2.0,
            2.0,
            4.0,
            4.0,
            2.0,
            2.25,
            1.25,
            -2.0,
            -4.0,
            -2.0,
            -2.0,
            -2.0,
        ]
        # Creating a basic docplex model
        mdl = Model("Test inequal")  # Docplex model
        num_z = 2  # Number of variables
        z = mdl.binary_var_list(num_z, name="Z")  # docplex variables
        objective = mdl.sum(z) - 2 * z[0] + z[1] * z[0] + 5  # objective function
        # Adding constraints
        mdl.add_constraint(mdl.sum(z[i] for i in range(num_z)) == 1)
        mdl.add_constraint(2 * z[0] + 3 * z[1] >= 1)
        mdl.add_constraint(2 * z[1] + z[0] <= 2)
        mdl.minimize(objective)  # Optimization

        ising_problem = FromDocplex2IsingModel(
            mdl
        ).ising_model  # Ising model of the Docplex model

        self.assertIsInstance(ising_problem, QUBO)
        self.assertEqual(ising_problem.weights, weights)

    def test_unbalanced_penalizations(self):
        """
        Test the equality and inequality constraints are encoded in the QUBO
        model using the unblanaced penalization method.

        """
        weights = [4.25, -0.95, -2.45]
        # Creating a basic docplex model
        mdl = Model("Test inequal")  # Docplex model
        num_z = 2  # Number of variables
        z = mdl.binary_var_list(num_z, name="Z")  # docplex variables
        objective = mdl.sum(z) - 2 * z[0] + z[1] * z[0] + 5  # objective function
        # Adding constraints
        mdl.add_constraint(mdl.sum(z[i] for i in range(num_z)) == 1)
        mdl.add_constraint(2 * z[0] + 3 * z[1] >= 1)
        mdl.add_constraint(2 * z[1] + z[0] <= 2)
        mdl.minimize(objective)  # Optimization

        ising_problem = FromDocplex2IsingModel(
            mdl, unbalanced_const=True
        ).ising_model  # Ising model of the Docplex model

        self.assertIsInstance(ising_problem, QUBO)

        for weight_1, weight_2 in zip(ising_problem.weights, weights):
            self.assertAlmostEqual(weight_1, weight_2)

    def test_model_maxcut(self):
        """
        Test the Maxcut application of OpenQAOA gives the same result as the
        Docplex translation model.
        """

        # Graph representing the maxcut problem
        n_nodes = 5  # number of nodes
        G = nx.Graph()
        G.add_nodes_from(range(n_nodes))
        G.add_edges_from([[0, 1], [0, 2], [1, 2], [4, 3], [3, 2], [4, 0], [2, 4]])

        # Docplex model
        mdl = Model(name="Max-cut")
        x = mdl.binary_var_list(n_nodes, name="x")
        for w, v in G.edges:
            G.edges[w, v].setdefault("weight", 1)
        objective = mdl.sum(
            G.edges[i, j]["weight"] * x[i] * (1 - x[j])
            + G.edges[i, j]["weight"] * x[j] * (1 - x[i])
            for i, j in G.edges
        )
        mdl.maximize(objective)

        # Translating the problem to OQ ising Model
        ModelOQ = FromDocplex2IsingModel(mdl)
        Ising_model_OQ = ModelOQ.ising_model.asdict()
        # Using the predefine function of this problem
        IsingModelDirect = MaximumCut(G).qubo.asdict()
        # Comparing both results in this specific case MaxCut from OpenQAOa gives
        # the coefficients omultiplied by two of the DocplexToIsingModel
        for nn, term in enumerate(IsingModelDirect["terms"]):
            idx_OQ = Ising_model_OQ["terms"].index(term)
            self.assertAlmostEqual(
                2 * Ising_model_OQ["weights"][idx_OQ], IsingModelDirect["weights"][nn]
            )


if __name__ == "__main__":
    unittest.main()



import networkx as nx
from docplex.mp.model import Model
from typing import Union
import matplotlib.pyplot as plt
from matplotlib import colormaps

from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class KColor(Problem):
    """
    Creates an instance of the k-color problem. It determines if given a set of
    colors, a graph can be colored such that no nodes connected by an edge share
    the same color.
    https://en.wikipedia.org/wiki/Graph_coloring
    Parameters
    ----------
    G: networkx.Graph
        Networkx graph of the problem
    k: int
        Maximum number of colors.
    penalty: float
        Penalty for the edge constraints.

    Returns
    -------
        An instance of the k-color problem.
    """

    __name__ = "k_color"

    def __init__(self, G, k, penalty: Union[int, float] = 2):
        self.G = G
        self.k = k
        self.penalty = penalty

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the k-color problem, whose graph is
        random following the Erdos-Renyi model.

        Parameters
        ----------
        n_nodes: int
            The number of nodes (vertices) in the graph.
        edge_probability: float
            The probability with which an edge is added to the graph.
        k: int
            Maximum number of colors.
        seed: int, optional
            Random seed for reproducibility.
        penalty: Union(int, float) Optional
            Penalty for the edge constraints.


        Returns
        -------
            A random instance of the k-color problem.
        """
        n_nodes = kwargs.get("n_nodes", 5)
        edge_probability = kwargs.get("edge_probability", 0.3)
        k = kwargs.get("k", 2)
        seed = kwargs.get("seed", None)
        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )
        return KColor(G, k)

    @property
    def docplex_model(self):
        mdl = Model("k_color")

        x = {
            (vertex, color): mdl.binary_var(name=f"x_{vertex}_{color}")
            for vertex in self.G
            for color in range(self.k)
        }

        # Constraint: Each vertex must be assigned to exactly one color
        for vertex in self.G:
            mdl.add_constraint(
                mdl.sum(x[(vertex, color)] for color in range(self.k)) == 1
            )
        cost = 0
        # Constraint: Adjacent vertices cannot have the same color
        for vertex1 in self.G:
            for vertex2 in self.G:
                if vertex2 in self.G[vertex1]:
                    for color in range(self.k):
                        cost += self.penalty * x[(vertex1, color)] * x[(vertex2, color)]
        # Objective function: Minimize the cost, finding a valid solution.
        mdl.minimize(cost)
        return mdl

    def classical_solution(self, string: bool = False):
        """
        Return the classical solution of the k-color problem

        Parameters
        ----------
        string : bool, optional
            If the solution is returned as a string. The default is False.

        Returns
        -------
        solution : Union[str, dict]
            The classical solution of the specific problem as a string or a dict.
        """
        mdl = self.docplex_model
        solution = mdl.solve()

        if solution is None:
            raise ValueError("No valid coloring found.")

        if string:
            coloring = ""
        else:
            coloring = {}
        for var in mdl.iter_binary_vars():
            if string:
                coloring += str(int(round(solution.get_value(var), 1)))
            else:
                coloring[var.name] = int(round(solution.get_value(var), 1))
        return coloring

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        qubo_docplex = FromDocplex2IsingModel(cplex_model).ising_model

        return QUBO(
            self.G.number_of_nodes(),
            qubo_docplex.terms + [[]],
            qubo_docplex.weights + [qubo_docplex.constant],
            self.problem_instance,
        )

    def plot_solution(self, solution: Union[dict, str], ax=None, colors=None):
        """
        A visualization method for the k-color solution.

        Parameters
        ----------
        solution : Union[dict, str]
            The solution of the specific k-color problem as a string or dictionary.
        ax : matplotlib axes, optional
            The default is None.

        Returns
        -------
        fig : matplotlib.pyplot.Figure()
            The graph visualization of the solution.
        """
        if isinstance(solution, str):
            sol = {}
            for n, var in enumerate(self.docplex_model.iter_binary_vars()):
                sol[var.name] = int(solution[n])
            solution = sol

        if ax is None:
            _, ax = plt.subplots()

        colors = colormaps["jet"] if colors is None else colors
        pos = nx.circular_layout(self.G)
        nx.draw(
            self.G,
            pos=pos,
            node_color=[
                colors((col + 1) / (self.k + 1))
                for vertex in self.G
                for col in range(self.k)
                if solution[f"x_{vertex}_{col}"] == 1
            ],
            edge_color="black",
            with_labels=True,
            ax=ax,
            alpha=0.85,
            **{"edgecolors": "black"},
        )
